import glob
import logging
import os
from threading import Thread
from typing import List

import foldedleastsquares.stats
import numpy
import matplotlib.pyplot as plt
import pandas as pd
from keras import backend as K
import keras
import numpy as np
import tensorflow as tf
import time

from astropy.timeseries import LombScargle
from keras.layers.activation.relu import ReLU
from keras.losses import BinaryCrossentropy
from keras.metrics import Precision, PrecisionAtRecall, Recall, AUC, BinaryAccuracy
from keras.optimizers import Adam
from keras.utils import plot_model
from keras.layers import LeakyReLU
from keras_nlp.src.layers import TransformerEncoder
from lightkurve import TessLightCurve
from numpy import ndarray
from sklearn.utils import shuffle

from exoml.ml.callback.early_stopping import ExoMlEarlyStopping
from exoml.ml.callback.learning_rate import WarmUpAndLinDecreaseCallback
from exoml.ml.callback.training_data_aware import ModelDirDataAwareCallback, ValidationDataAwareCallback
from exoml.ml.log.get_weights_logger import ModelWeightsLogger
from exoml.ml.log.with_logging import WithLogging
from exoml.ml.model.base_model import HyperParams
from exoml.santo.santo_global_generator import SantoGlobalGenerator
from exoml.santo.santo_global_predict_generator import SantoPredictGenerator


def clean_to_float(x):
    return float(str(x).strip("\"'").strip("\"").strip("'"))


def weighted_binary_crossentropy_param(weight=1):
    def weighted_binary_crossentropy(y_true, y_pred, weight=weight):
        y_true = K.clip(y_true, K.epsilon(), 1 - K.epsilon())
        y_pred = K.clip(y_pred, K.epsilon(), 1 - K.epsilon())
        logloss = -(y_true * K.log(y_pred) * weight + (1 - y_true) * K.log(1 - y_pred))
        return K.mean(logloss, axis=-1)
    return weighted_binary_crossentropy


class SANTOGlobal(WithLogging):
    """
    Self-Attention Neural Network for Transiting Objects
    """
    def __init__(self, transformer_layers=0, transformer_heads=1, name="SANTO") -> None:
        super().__init__()
        self.name = "SANTO"
        self.steps_per_epoch = 1
        self.transformer_layers = transformer_layers
        self.transformer_heads = transformer_heads

    def load(self, dir, hyperparams: HyperParams):
        self.model = tf.keras.models.load_model(dir) #, custom_objects={'loss': self.loss(hyperparams)})


    def inform(self, dir):
        """
        Creates summary and a visual plot of the model
        :param dir: the directory to store the plot
        :return: the object itself
        """
        logging.info("Creating model plot and summary")
        dir = dir + self._get_model_dir()
        if not os.path.exists(dir):
            os.mkdir(dir)
        try:
            plot_model(self.model, dir + '/network.png', show_shapes=True)
        except Exception as e:
            logging.exception("Can't plot model network.png")
        self.model.summary()
        ModelWeightsLogger().log_model_weights(self.model)
        return self

    def initialize(self, lcs_dir, output_dir, hyperparams: HyperParams, input_dim=500,
              cores=os.cpu_count() // 2, load_checkpoint=False):
        input_layer = keras.Input(shape=(input_dim, 1), name="input")
        if self.transformer_layers > 0:
            # outputs = TransformerClassifier(transformer_input_size=input_dim, patch_size=input_dim,
            #                                 num_heads=self.transformer_heads, mlp_dim=input_dim,
            #                                 hyperparams=hyperparams, num_blocks=self.transformer_layers,
            #                                 classes=1)(input_layer)
            flux_branch: keras.layers.Layer = input_layer
            for _ in range(self.transformer_layers):
                flux_branch = TransformerEncoder(
                    intermediate_dim=input_dim,
                    num_heads=self.transformer_heads,
                    dropout=hyperparams.dropout_rate,
                    activation=LeakyReLU(),
                    layer_norm_epsilon=1e-05,
                    kernel_initializer="glorot_uniform",
                    bias_initializer="zeros",
                    normalize_first=False
                )(flux_branch)
                # flux_branch = keras.layers.MultiHeadAttention(
                #     key_dim=input_dim, num_heads=self.transformer_heads, dropout=hyperparams.dropout_rate
                # )(input_layer, input_layer)
                # flux_branch = keras.layers.Dropout(hyperparams.dropout_rate)(flux_branch)
                # flux_branch = keras.layers.LayerNormalization(epsilon=1e-6)(flux_branch)
                # res = flux_branch + input_layer
                # # Feed Forward Part
                # flux_branch = keras.layers.Conv1D(filters=input_dim, kernel_size=1, activation="relu")(res)
                # flux_branch = keras.layers.Dropout(hyperparams.dropout_rate)(flux_branch)
                # flux_branch = keras.layers.Conv1D(filters=input_layer.shape[-1], kernel_size=1)(flux_branch)
                # flux_branch = keras.layers.LayerNormalization(epsilon=1e-6)(flux_branch)
                # flux_branch = flux_branch + res
            flux_branch = keras.layers.GlobalAveragePooling1D(data_format="channels_last")(flux_branch)
            flux_branch = keras.layers.Dense(input_dim, activation="relu")(flux_branch)
            flux_branch = keras.layers.Dropout(hyperparams.dropout_rate)(flux_branch)
            outputs = keras.layers.Dense(1, activation="sigmoid")(flux_branch)
        else:
            leaky_relu_alpha = 0.01
            flux_branch: keras.layers.Layer = input_layer
            # flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
            #flux_branch = keras.layers.Dense(64, activation=LeakyReLU())(keras.layers.Flatten()(flux_branch))
            conv_small = keras.layers.Conv1D(filters=16, kernel_size=3, padding="same",
                                             activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
            conv_medium = keras.layers.Conv1D(filters=16, kernel_size=15, padding="same",
                                 activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
            conv_large = keras.layers.Conv1D(filters=16, kernel_size=90, padding="same",
                                activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
            flux_branch = keras.layers.Concatenate(axis=-1)([conv_small, conv_medium, conv_large])
            flux_branch = keras.layers.SpatialDropout1D(hyperparams.dropout_rate)(flux_branch)
            flux_branch = keras.layers.Conv1D(filters=16, kernel_size=90, padding="same",
                                              activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
            flux_branch = keras.layers.MaxPooling1D(pool_size=10, strides=2)(flux_branch)
            # flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
            flux_branch = keras.layers.SpatialDropout1D(rate=hyperparams.spatial_dropout_rate)(flux_branch)
            conv1_branch = flux_branch
            conv1_branch: keras.layers.Layer = keras.layers.Flatten()(conv1_branch)
            conv1_branch = keras.layers.Dense(8, activation=LeakyReLU(leaky_relu_alpha),
                                              kernel_regularizer=tf.keras.regularizers.L1L2(l1=hyperparams.l1_regularization, l2=hyperparams.l2_regularization))(conv1_branch)
            conv1_branch = keras.layers.Dropout(hyperparams.dropout_rate)(conv1_branch)
            flux_branch = keras.layers.Conv1D(filters=32, kernel_size=30, padding="same",
                                              activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
            flux_branch = keras.layers.MaxPooling1D(pool_size=4, strides=2)(flux_branch)
            # flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
            flux_branch = keras.layers.SpatialDropout1D(rate=hyperparams.spatial_dropout_rate)(flux_branch)
            conv2_branch = flux_branch
            conv2_branch: keras.layers.Layer = keras.layers.Flatten()(conv2_branch)
            conv2_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha),
                                              kernel_regularizer=tf.keras.regularizers.L1L2(l1=hyperparams.l1_regularization, l2=hyperparams.l2_regularization))(conv2_branch)
            conv2_branch = keras.layers.Dropout(hyperparams.dropout_rate)(conv2_branch)
            flux_branch = keras.layers.Conv1D(filters=64, kernel_size=15, padding="same",
                                              activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
            flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=2)(flux_branch)
            # flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
            flux_branch = keras.layers.SpatialDropout1D(rate=hyperparams.spatial_dropout_rate)(flux_branch)
            conv3_branch = flux_branch
            conv3_branch: keras.layers.Layer = keras.layers.Flatten()(conv3_branch)
            conv3_branch = keras.layers.Dense(32, activation=LeakyReLU(leaky_relu_alpha),
                                              kernel_regularizer=tf.keras.regularizers.L1L2(l1=hyperparams.l1_regularization, l2=hyperparams.l2_regularization))(conv3_branch)
            conv3_branch = keras.layers.Dropout(hyperparams.dropout_rate)(conv3_branch)
            flux_branch = keras.layers.Conv1D(filters=128, kernel_size=8, padding="same",
                                              activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
            flux_branch = keras.layers.MaxPooling1D(pool_size=2, strides=2)(flux_branch)
            conv4_branch = flux_branch
            conv4_branch: keras.layers.Layer = keras.layers.Flatten()(conv4_branch)
            conv4_branch = keras.layers.Dense(64, activation=LeakyReLU(leaky_relu_alpha),
                                              kernel_regularizer=tf.keras.regularizers.L1L2(l1=hyperparams.l1_regularization, l2=hyperparams.l2_regularization))(conv4_branch)
            conv4_branch = keras.layers.Dropout(hyperparams.dropout_rate)(conv4_branch)
            flux_branch = keras.layers.Concatenate(axis=1)([
                conv1_branch, conv2_branch, conv3_branch, conv4_branch])
            flux_branch = keras.layers.Dense(512, activation=LeakyReLU(leaky_relu_alpha),
                              kernel_regularizer=tf.keras.regularizers.L1L2(l1=hyperparams.l1_regularization, l2=hyperparams.l2_regularization))(flux_branch)
            # flux_branch = keras.layers.Dropout(hyperparams.dropout_rate)(flux_branch)
            # flux_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha),
            #                   kernel_regularizer=tf.keras.regularizers.L1L2(l1=hyperparams.l1_regularization, l2=hyperparams.l2_regularization))(flux_branch)
            flux_branch = keras.layers.Dropout(hyperparams.dropout_rate)(flux_branch)
            outputs: keras.layers.Layer = keras.layers.Dense(input_dim, activation="sigmoid")(flux_branch)
        print("Preparing data sets")
        self.model: keras.Model = keras.Model(inputs=input_layer, outputs=outputs)
        self.model.run_eagerly = hyperparams.run_eagerly
        model_path = output_dir + self._get_model_dir()
        train_filenames, validation_filenames, test_filenames = \
            self.prepare_training_data(lcs_dir, model_path, hyperparams.batch_size, hyperparams.train_percent,
                                       hyperparams.validation_percent, hyperparams.training_set_limit,
                                       hyperparams.balance_class_id, hyperparams.balance_class_sampling)

    def train(self, lcs_dir, output_dir, hyperparams: HyperParams, input_dim=500, step_indexes=1):
        print("Begin training")
        tf.keras.backend.clear_session()
        self.inform(output_dir)
        model_path = output_dir + self._get_model_dir()
        if not os.path.exists(model_path):
            os.mkdir(model_path)
        train_df = pd.read_csv(model_path + '/train_dataset.csv')
        train_filenames = train_df.loc[:, 'name'].tolist()
        validation_df = pd.read_csv(model_path + '/validation_dataset.csv')
        validation_filenames = validation_df.loc[:, 'name'].tolist()
        training_batch_generator = self.instance_generator(train_filenames, lcs_dir, hyperparams.batch_size,
                                                           input_dim, None, hyperparams.zero_epsilon,
                                                           shuffle=False, step_indexes=step_indexes)
        steps_per_epoch = training_batch_generator.steps_per_epoch()
        optimizer: keras.optimizers.Optimizer = self.build_optimizer(hyperparams, steps_per_epoch)
        loss, accuracy = self.instance_loss_accuracy(hyperparams)
        self.compile(optimizer, loss,
                     metrics=[accuracy] + self.instance_metrics() + hyperparams.metrics,
                     run_eagerly=hyperparams.run_eagerly)
        if not hyperparams.dry_run:
            logging.info("Initializing training and validation generators with (batch_size," + str(hyperparams.batch_size) +
                         ") (self.input_size," + str(input_dim) +
                         ") (zero_epsilon," + str(hyperparams.zero_epsilon) +
                         ")")
            validation_batch_generator = self.instance_generator(validation_filenames, lcs_dir, hyperparams.batch_size,
                                                                 input_dim, None, hyperparams.zero_epsilon,
                                                                 shuffle=False, step_indexes=step_indexes)
            for callback in hyperparams.callbacks:
                if issubclass(callback.__class__, ModelDirDataAwareCallback):
                    callback.set_model_dir(model_path)
                if issubclass(callback.__class__, ValidationDataAwareCallback):
                    callback.set_validation_data(validation_batch_generator)
            if hyperparams.early_stopping_patience > 0 and hyperparams.early_stopping_delta > 0:
                hyperparams.callbacks = hyperparams.callbacks + [ExoMlEarlyStopping(
                    monitor="val_loss",
                    min_delta=hyperparams.early_stopping_delta,
                    patience=hyperparams.early_stopping_patience,
                    verbose=0,
                    mode="auto",
                    baseline=None,
                    restore_best_weights=False,
                )]
            model_validation_steps = validation_batch_generator.steps_per_epoch()
            class_weights = hyperparams.class_loss_weights if hyperparams.class_loss_weights is not None \
                else training_batch_generator.class_weights()
            logging.info("Initializing training with (epochs," + str(hyperparams.epochs) +
                         ") (steps_per_epoch," + str(steps_per_epoch) +
                         ")")
            self.save(output_dir)
            fit_history = self.fit_model(hyperparams, training_batch_generator,
                                         steps_per_epoch, class_weights,
                                         validation_batch_generator,
                                         model_validation_steps, hyperparams.epochs)
            self.save(output_dir)

    def predict(self, lcs_dir: str, target_files: list, model_dir: str, batch_size: int = 1024, input_size: int = 2500,
                zero_epsilon: float = 1e-7, plot=False, plot_positives=False, tagged_data: bool = False,
                target_name: str = ""):
        self.load_model(model_dir)
        half_window = input_size // 2
        generator = SantoPredictGenerator(lcs_dir, target_files, input_size=input_size, step_size=1, batch_size=batch_size, zero_epsilon=zero_epsilon)
        predictions = self.model.predict(generator)
        left = np.linspace(0.1, 1.0, input_size // 2, endpoint=False)
        right = np.linspace(1.0, 0.1, input_size // 2)
        triangle_weights = np.concatenate([left, right])
        current_index = 0
        files_predictions = {}
        for target_file in target_files:
            file_content = pd.read_csv(lcs_dir + '/' + target_file).values
            flux_key = 1
            if file_content.shape[0] > 40:
                file_content = np.transpose(file_content)
                flux_key = 2
            file_content = np.vectorize(clean_to_float)(file_content)
            file_predictions = predictions[current_index:current_index + file_content.shape[1]]
            N = file_predictions.shape[0]
            final_preds = np.zeros(N)
            weight_sums = np.zeros(N)
            for i in range(N):
                j_start = max(0, i - half_window)
                j_end = min(N, i + half_window)
                # corresponding columns in the prediction matrix
                js = np.arange(j_start, j_end)
                ks = half_window - (js - i)
                # Remove invalid ks (outside [0, 2499])
                valid_mask = (ks >= 0) & (ks < input_size)
                js = js[valid_mask]
                ks = ks[valid_mask]
                pred_values = predictions[js, ks]
                weights = triangle_weights[half_window - (js - i)]
                final_preds[i] = np.sum(pred_values * weights)
                weight_sums[i] = np.sum(weights)
            final_preds /= weight_sums
            files_predictions[target_file] = final_preds
            if plot:
                fig, axs = plt.subplots(2 if not tagged_data else 3, 1, figsize=(16, 8),
                                        constrained_layout=True)
                axs[0].scatter(np.arange(0, N), file_content[flux_key][0:N])
                axs[1].plot(np.arange(0, N), final_preds)
                if tagged_data:
                    axs[2].plot(np.arange(0, N), file_content[2][0:N])
                plt.title(target_file)
                plt.show()
            current_index = current_index + len(file_content)
        return files_predictions

    def tokenize_found_transits(self, time, flux, predictions, window_size=500, plot=False, tagged_data=None):
        positive_predictions = predictions
        for _ in np.arange(0, 10):
            positive_predictions = np.convolve(positive_predictions, np.ones(25) / 25, mode='same')
        sigma_low_indexes = np.argwhere(predictions < 3 * np.nanstd(predictions)).flatten()
        positive_predictions[sigma_low_indexes] = 0
        diff = np.diff(positive_predictions)
        local_maxima = (np.hstack([diff, 0]) < 0) & (np.hstack([0, diff]) > 0)
        local_maxima_indexes = np.argwhere(local_maxima).flatten()
        if plot:
            for index in local_maxima_indexes:
                if index > window_size // 2:
                    fig, axs = plt.subplots(2 if tagged_data is None else 3, 1, figsize=(16, 8), constrained_layout=True)
                    above_indexes = np.argwhere(predictions[index - window_size // 2:index + window_size // 2] >= 3 * np.nanstd(predictions))
                    below_indexes = np.argwhere(predictions[index - window_size // 2:index + window_size // 2] < 3 * np.nanstd(predictions))
                    axs[0].scatter(time[index - window_size // 2:index + window_size // 2][above_indexes],
                                   flux[index - window_size // 2:index + window_size // 2][above_indexes],
                                   color='blue')
                    axs[1].scatter(time[index - window_size // 2:index + window_size // 2][above_indexes],
                                predictions[index - window_size // 2:index + window_size // 2][above_indexes],
                                color='blue')
                    axs[0].scatter(time[index - window_size // 2:index + window_size // 2][below_indexes],
                                   flux[index - window_size // 2:index + window_size // 2][below_indexes],
                                   color='firebrick')
                    axs[1].scatter(time[index - window_size // 2:index + window_size // 2][below_indexes],
                                predictions[index - window_size // 2:index + window_size // 2][below_indexes],
                                color='firebrick')
                    if tagged_data is not None:
                        axs[2].scatter(time[index - window_size // 2:index + window_size // 2][above_indexes],
                                    tagged_data[index - window_size // 2:index + window_size // 2][above_indexes],
                                    color='blue')
                    plt.show()
                    plt.close(fig)
                    plt.clf()
        return local_maxima_indexes


    def load_model(self, dir, custom_objects={}, compile=False):
        custom_objects['WarmUpAndLinDecreaseCallback'] = WarmUpAndLinDecreaseCallback
        custom_objects['ThresholdAtPrecision'] = None
        logging.info("Loading model from %s", dir)
        self.set_model(tf.keras.models.load_model(dir, compile=compile, custom_objects=custom_objects))
        return self

    def set_model(self, model):
        """
        Stores the model in a class attribute
        :param model: the model to be stored
        """
        self.model = model

    def _get_model_dir(self):
        return self.name + '_model/'

    def instance_loss_accuracy(self, hyperparams: HyperParams):
        return BinaryCrossentropy() if hyperparams.custom_loss is None else hyperparams.custom_loss, BinaryAccuracy()

    def compile(self, optimizer, loss, metrics=None, run_eagerly=False, tuner=False):
        """
        Compiles and prepares the model for training
        :param optimizer: the optimizer to be used
        :param loss: the loss to be used
        :param metrics: the metrics to be used
        :return: the object itself
        """
        if metrics is None:
            metrics = []
        metrics_str = [str(metric) for metric in metrics]
        logging.info("Compiling model with optimizer " + str(optimizer) + ", loss " + str(loss) + " and metrics [" +
                     ",".join(metrics_str) + "]")
        self.model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
        self.model.run_eagerly = run_eagerly
        return self

    def instance_metrics(self):
        return [Precision(name="precision"), Recall(name="recall"),
                #F1Score(num_classes=1, threshold=0.5, average='weighted', name='f1_score'),
                PrecisionAtRecall(recall=0.99, name="p@r99", num_thresholds=1000),
                #ThresholdAtPrecision(precision=0.99, name="t@p99", num_thresholds=1000),
                AUC(name="roc_auc"),
                AUC(curve="PR", name="pr_auc")]

    def instance_generator(self, dataset, dir, batch_size, input_sizes, type_to_label, zero_epsilon, shuffle=True,
                           step_indexes=1):
        return SantoGlobalGenerator(dir, dataset, input_size=input_sizes, step_size=1, batch_size=batch_size,
                              shuffle=shuffle, zero_epsilon=zero_epsilon, indexes_steps=step_indexes)

    def load_training_set(self, **kwargs):
        training_dir = kwargs.get('training_dir')
        target_files = glob.glob(f'/{training_dir}/**/*.csv', recursive=True)
        target_files = [os.path.basename(path) for path in target_files]
        return target_files

    def prepare_training_data(self, training_dir, output_dir, batch_size, train_percent=0.8, validation_percent=0.1,
                              training_set_limit=None, balance_class_id=None, balance_class_sampling=None):
        training_dataset = self.load_training_set(training_dir=training_dir)
        training_dataset = shuffle(training_dataset)
        test_dataset = None
        training_dataset = shuffle(training_dataset)
        if training_set_limit is not None:
            if isinstance(training_dataset, pd.DataFrame):
                training_dataset = training_dataset[:training_set_limit]
            else:
                training_dataset = training_dataset[:training_set_limit]
        dataset_length = len(training_dataset)
        train_last_index = int(dataset_length * train_percent)
        validation_last_index = train_last_index + int(dataset_length * validation_percent)
        validation_last_index = validation_last_index if validation_last_index < dataset_length else dataset_length
        test_last_index = dataset_length - 1
        train_dataset_filename = output_dir + "/train_dataset.csv"
        validation_dataset_filename = output_dir + "/validation_dataset.csv"
        test_dataset_filename = output_dir + "/test_dataset.csv"
        logging.info("Storing train and test file names in " + train_dataset_filename + " and " +
                     validation_dataset_filename)
        if os.path.exists(train_dataset_filename):
            os.remove(train_dataset_filename)
        if os.path.exists(validation_dataset_filename):
            os.remove(validation_dataset_filename)
        if os.path.exists(test_dataset_filename):
            os.remove(test_dataset_filename)
        train_filenames = training_dataset[0:train_last_index]
        validation_filenames = training_dataset[train_last_index:validation_last_index]
        test_filenames = []
        if validation_last_index != test_last_index:
            test_filenames = test_dataset if test_dataset is not None else training_dataset[validation_last_index:test_last_index]
            test_df = pd.DataFrame(columns=[])
            test_df['name'] = test_filenames
            test_df.to_csv(test_dataset_filename)
        train_df = pd.DataFrame(columns=['name'])
        train_df['name'] = train_filenames
        val_df = pd.DataFrame(columns=['name'])
        val_df['name'] = validation_filenames
        train_df.to_csv(train_dataset_filename)
        val_df.to_csv(validation_dataset_filename)
        logging.info("Training set is of length %s (%s %%)", len(train_filenames),
                     len(train_filenames) / dataset_length * 100)
        logging.info("Validation set is of length %s (%s %%)", len(validation_filenames),
                     len(validation_filenames) / dataset_length * 100)
        logging.info("Testing set is of length %s (%s %%)", len(test_filenames),
                     len(test_filenames) / dataset_length * 100)
        return train_filenames, validation_filenames, test_filenames

    def save(self, dir, model_dir=None):
        if model_dir is None:
            dest_dir = dir + '/' + self._get_model_dir()
        else:
            dest_dir = dir + '/' + model_dir
        logging.info("Saving model into %s", dest_dir)
        self.model.save(dest_dir)
        logging.info("Saved model")
        return dest_dir

    def fit_model(self, hyperparams, training_batch_generator, steps_per_epoch, class_weights,
                  validation_batch_generator, model_validation_steps, epochs):
        fit_history = self.model.fit(x=training_batch_generator,
                                     steps_per_epoch=steps_per_epoch,
                                     epochs=hyperparams.epochs, verbose=1,
                                     validation_data=validation_batch_generator,
                                     validation_steps=model_validation_steps,
                                     callbacks=hyperparams.callbacks,
                                     use_multiprocessing=hyperparams.cores > 0, workers=1
            if hyperparams.cores <= 0 else hyperparams.cores)
        return fit_history

    def compute_initial_lr(self, hyperparams: HyperParams):
        if hyperparams.lr_progression != 1:
            initial_lr = hyperparams.initial_learning_rate
        else:
            initial_lr = hyperparams.learning_rate_schedule \
                if hyperparams.learning_rate_schedule is not None else hyperparams.initial_learning_rate
        return initial_lr

    def build_optimizer(self, hyperparams, steps_per_epoch):
        # from exoml.ml.callback.learning_rate import MultiOptimizer
        from tensorflow_addons.optimizers import MultiOptimizer
        initial_lr = self.compute_initial_lr(hyperparams)
        optimizer = self.build_swa_optimizer(hyperparams, initial_lr, steps_per_epoch)
        if hyperparams.lr_progression != 1:
            optimizer.progressive_lr_factor = 1
            optimizers_and_layers = []
            standard_layers = []
            progressive_lr_factor = 1
            for layer in self.model.layers:
                if 'final' in layer.name:
                    progressive_lr_factor = progressive_lr_factor * hyperparams.lr_progression
                    progressive_optimizer = self.build_swa_optimizer(hyperparams, progressive_lr_factor * initial_lr, steps_per_epoch)
                    progressive_optimizer.progressive_lr_factor = progressive_lr_factor
                    optimizers_and_layers = optimizers_and_layers + [
                        (progressive_optimizer, layer)]
                else:
                    standard_layers = standard_layers + [layer]
            optimizers_and_layers = [(optimizer, standard_layers)] + optimizers_and_layers
            optimizer = MultiOptimizer(optimizers_and_layers)
        return optimizer

    def build_swa_optimizer(self, hyperparams: HyperParams, lr, steps_per_epoch):
        # from exoml.ml.callback.learning_rate import SWA
        from tensorflow_addons.optimizers import SWA
        if hyperparams.stochastic_weight_average_wait > 0:
            optimizer = tf.keras.optimizers.legacy.Adam(lr,
                                                        beta_1=0.9, beta_2=0.98, epsilon=1e-9,
                                                        clipnorm=hyperparams.gradient_clip_norm,
                                                        clipvalue=hyperparams.gradient_clip_value)
            optimizer = SWA(optimizer,
                                           start_averaging=steps_per_epoch * hyperparams.stochastic_weight_average_wait,
                                           average_period=steps_per_epoch)
            if isinstance(lr, (int, float)):
                optimizer.lr = lr
            else:
                optimizer.lr = hyperparams.initial_learning_rate
        else:
            optimizer = Adam(lr,
                                                 beta_1=0.9, beta_2=0.98, epsilon=1e-9,
                                                 clipnorm=hyperparams.gradient_clip_norm,
                                                 clipvalue=hyperparams.gradient_clip_value)
        return optimizer

    def run_epoch(self, epoch, lcs_dir, model_dir, target_files, hyperparams, input_dim, transformer_model, optimizer,
                  training=False, cores=os.cpu_count() // 2, training_log_df=None):
        training_preparation_threads: list[Thread] = []
        losses = []
        accuracies = []
        precisions = []
        recalls = []
        pr99s = []
        accuracy: keras.metrics.Metric = tf.keras.metrics.SparseCategoricalAccuracy(name="accuracy")
        precision: keras.metrics.Metric = tf.keras.metrics.Precision(name="precision")
        recall: keras.metrics.Metric = tf.keras.metrics.Recall(name="recall")
        pr99: keras.metrics.Metric = tf.keras.metrics.PrecisionAtRecall(recall=0.99, name="p@r99")
        step = 0
        for file_index, last_target_file in enumerate(target_files):
            last_target_file_pos: int = 0
            flux = np.loadtxt(f'{lcs_dir}/{last_target_file}', delimiter=',')
            if np.nanmin(flux[1]) == 1:
                continue
            train_batch: ndarray = np.zeros((hyperparams.batch_size, input_dim))
            train_tags: ndarray = np.zeros((hyperparams.batch_size, 1))
            iteration: int = 0
            data_remaining: bool = True
            while data_remaining:
                # training_preparation_threads: list[Thread] = []
                # for core in range(0, cores):
                #     training_preparation_threads =
                batch_index = iteration // hyperparams.batch_size
                iteration_index = iteration % hyperparams.batch_size
                # TODO mask all out of transit information not between 3 durations of a transit
                flux_data = flux[0][last_target_file_pos:last_target_file_pos + input_dim] / 2
                tags_data_index = last_target_file_pos + input_dim // 2 - 1
                tags_data = flux[1][tags_data_index - input_dim:tags_data_index + input_dim]
                last_target_file_pos = last_target_file_pos + 1
                data_remaining = last_target_file_pos < flux.shape[1] - input_dim
                is_last_iteration = (iteration + 1) % hyperparams.batch_size == 0 or not data_remaining
                if not training or is_last_iteration or np.any(np.abs(tags_data - 1) > 1e-6):
                    tags_data = flux[1][tags_data_index]
                    train_batch[iteration_index] = flux_data
                    train_tags[iteration_index] = tags_data
                    oot_mask = np.abs(train_tags[iteration_index] - 1) < 1e-6
                    train_tags[iteration_index][oot_mask] = 0
                    train_tags[iteration_index][~oot_mask] = 1
                    if is_last_iteration:
                        #oot_mask = np.abs(train_tags.flatten()) < 1e-6
                        #plt.scatter(np.arange(0, iteration_index + 1)[oot_mask], flux[0][0:iteration_index + 1][oot_mask],
                        #            color='blue')
                        #plt.scatter(np.arange(0, iteration_index + 1)[~oot_mask], flux[0][0:iteration_index + 1][~oot_mask],
                        #            color='red')
                        #plt.show()
                        #positives_positions = np.argwhere(train_tags.flatten() == True)
                        #plt.scatter(np.arange(0, 500), train_batch[positives_positions][0], color = 'blue')
                        #plt.show()
                        train_tags_tensor = tf.convert_to_tensor(train_tags, dtype=np.float32)
                        train_batch_tensor = tf.convert_to_tensor(train_batch, dtype=np.float32)
                        train_tags_tensor = tf.reshape(train_tags_tensor, [hyperparams.batch_size, 1])
                        if training:
                            with tf.GradientTape() as tape:
                                tape.watch(transformer_model.trainable_variables)
                                flux_predictions = transformer_model(train_batch)
                                loss = transformer_model.compute_loss(train_batch_tensor, train_tags_tensor, flux_predictions)
                            losses.append(loss.numpy())
                            gradients = tape.gradient(loss, transformer_model.trainable_variables)
                            optimizer.apply_gradients(zip(gradients, transformer_model.trainable_variables))
                            for gradient in gradients:
                                if np.any(gradient.numpy() == 0):
                                    print("There are gradients with zero values")
                                    break
                            if np.all(flux_predictions == 0):
                                print("All predictions are 0")
                            if np.all(flux_predictions == 1):
                                print("All predictions are 1")
                            step = step + 1
                        else:
                            flux_predictions = transformer_model.predict_on_batch(train_batch)
                        # flux_predictions_reshape = tf.reshape(flux_predictions, [hyperparams.batch_size, input_dim])
                        train_tags_tensor = tf.reshape(train_tags_tensor, [hyperparams.batch_size, 1])
                        accuracy(train_tags_tensor, flux_predictions)
                        precision(train_tags_tensor, flux_predictions)
                        recall(train_tags_tensor, flux_predictions)
                        pr99(train_tags_tensor, flux_predictions)
                        accuracies.append(accuracy.result().numpy())
                        precisions.append(precision.result().numpy())
                        recalls.append(recall.result().numpy())
                        pr99s.append(pr99.result().numpy())
                        if training:
                            print("Epoch {} File {} Batch {} Loss {} Accuracy {} Precision {} Recall {} P@R99 {}"
                                  .format(epoch + 1, file_index, step, loss, accuracy.result(),
                                          precision.result(), recall.result(), pr99.result()))
                            if training_log_df is not None:
                                if isinstance(self.model.optimizer, MultiOptimizer):
                                    learning_rate = K.eval(K.eval(self.model.optimizer.optimizer_specs[0]['optimizer'].lr))
                                else:
                                    learning_rate = K.eval(self.model.optimizer._decayed_lr(tf.float32))
                                training_log_df = pd.concat([training_log_df, pd.DataFrame.from_dict(
                                    {'lr': [learning_rate], 'epoch': [epoch], 'batch': [step], 'loss': [loss.numpy()],
                                     'accuracy': [accuracy.result().numpy()], 'precision': [precision.result().numpy()],
                                     'recall': [recall.result().numpy()],
                                     'p@r99': [pr99.result().numpy()]})], ignore_index=True)
                                training_log_df.to_csv(model_dir + '/training_log.csv', index=False)
                                #self.plot_metrics(model_dir, self.steps_per_epoch, False)
                            accuracy.reset_state()
                            precision.reset_state()
                            recall.reset_state()
                            pr99.reset_state()
                    iteration = iteration + 1
        if not training: 
            print(f"Validation Epoch {epoch + 1} Accuracy {accuracy.result().numpy()} "
                  f"Precision {precision.result().numpy()} Recall {recall.result().numpy()} P@R99 "
                  f"{pr99.result().numpy()}")
            if training_log_df is not None:
                if isinstance(self.model.optimizer, MultiOptimizer):
                    learning_rate = K.eval(K.eval(self.model.optimizer.optimizer_specs[0]['optimizer'].lr))
                else:
                    learning_rate = K.eval(self.model.optimizer._decayed_lr(tf.float32))
                training_log_df = pd.concat([training_log_df, pd.DataFrame.from_dict(
                    {'lr': [learning_rate], 'epoch': [epoch], 'batch': [0],
                     'val_accuracy': [accuracy.result().numpy()], 'val_precision': [precision.result().numpy()],
                     'val_recall': [recall.result().numpy()], 'val_p@r99': [pr99.result().numpy()]})], ignore_index=True)
                training_log_df.to_csv(model_dir + '/training_log.csv', index=False)
        return losses, accuracies, precisions, recalls, pr99s, training_log_df

    def plot_metrics(self, model_dir, steps_per_epoch, training_metrics=[], validation_metrics=[]):
        logging.info("Plotting metrics")
        training_log_df = pd.read_csv(model_dir + '/training_log.csv')
        training_log_df['batch_total'] = training_log_df['batch'] + (training_log_df['epoch'] * steps_per_epoch)
        validation_log_df = training_log_df.dropna(subset=['val_precision', 'val_recall', 'val_p@r99', 'val_accuracy'])
        fig, axs = plt.subplots(2, 2, figsize=(25, 7), constrained_layout=True)
        fig.suptitle(self.name + ' Model Metrics', fontsize=14)
        for metric in validation_metrics:
            axs[metric[2][0]][metric[2][1]].plot(validation_log_df['batch_total'], validation_log_df['val_' + metric[0]],
                           color=metric[1], label='val_' + metric[0], linestyle='--')
        for metric in training_metrics:
            axs[metric[2][0]][metric[2][1]].plot(validation_log_df['batch_total'], validation_log_df[metric[0]],
                           color=metric[1], label=metric[0])
        axs[0][0].set_xlabel("Batch")
        secx = axs[0][0].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[0][0].set_ylabel("Learning rate")
        axs[0][0].legend(loc='upper right')
        axs[0][1].set_xlabel("Batch")
        secx = axs[0][1].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[0][1].set_ylabel("Loss")
        axs[0][1].legend(loc='upper right')
        axs[1][0].set_xlabel("Batch")
        secx = axs[1][0].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[1][0].set_ylabel("Precision/Recall/P@R")
        axs[1][0].legend(loc='upper left')
        axs[1][1].set_xlabel("Batch")
        secx = axs[1][1].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[1][1].set_ylabel("Accuracy/F1/AUCs")
        axs[1][1].legend(loc='upper left')
        plt.savefig(model_dir + '/metrics.png')
        plt.close(fig)
        plt.clf()
        logging.info("Plotted metrics")

# SANTO().train("/data/scratch/ml/santo/training_data/", "/data/scratch/exoml/SANTO",
#               hyperparams=HyperParams(batch_size=100, epochs=10, run_eagerly=True))



import logging

from keras.losses import BinaryCrossentropy
from keras.metrics import Precision, Recall, AUC, BinaryAccuracy, RecallAtPrecision
from exoml.ml.loss.cross_entropy import binary_focal_loss
from exoml.ml.metrics.auc import precision_at_k, mean_true_positive_value, mean_true_negative_value, \
    mean_false_negative_value, mean_false_positive_value, ThresholdAtPrecision
from exoml.ml.model.base_model import BaseModel
import pandas as pd
import matplotlib.pyplot as plt


class BinaryModel(BaseModel):
    def __init__(self, name, input_size, class_ids, type_to_label, hyperparams) -> None:
        super().__init__(name, input_size, class_ids, type_to_label, hyperparams)

    def instance_loss_accuracy(self):
        return BinaryCrossentropy(), BinaryAccuracy()

    def instance_metrics(self):
        return [Precision(name="precision"), Recall(name="recall"),
                #F1Score(num_classes=1, threshold=0.5, average='weighted', name='f1_score'),
                RecallAtPrecision(precision=0.99, name="r@p99", num_thresholds=1000),
                RecallAtPrecision(precision=1.0, name="r@p100", num_thresholds=1000),
                #ThresholdAtPrecision(precision=0.99, name="t@p99", num_thresholds=1000),
                AUC(name="roc_auc"),
                AUC(curve="PR", name="pr_auc")]

    def plot_metrics(self, model_dir, steps_per_epoch, with_validation=False):
        logging.info("Plotting metrics")
        training_log_df = pd.read_csv(model_dir + '/training_log.csv')
        training_log_df['batch_total'] = training_log_df['batch'] + (training_log_df['epoch'] * steps_per_epoch)
        validation_log_df = training_log_df.dropna(subset = ['val_precision', 'val_recall', 'val_r@p99', 'val_loss', 'val_binary_accuracy',
                                                             'val_pr_auc', 'val_roc_auc'])
        fig, axs = plt.subplots(2, 2, figsize=(25, 7), constrained_layout=True)
        fig.suptitle(self.name + ' Model Metrics', fontsize=14)
        axs[0][0].plot(training_log_df['batch_total'], training_log_df['lr'])
        axs[0][1].plot(training_log_df['batch_total'], training_log_df['loss'], color='blue', label='Train loss')
        if with_validation:
            axs[0][1].plot(validation_log_df['epoch'] * steps_per_epoch, validation_log_df['val_loss'], color='red',
                           label='Val. loss')
        axs[1][0].plot(training_log_df['batch_total'], training_log_df['precision'], color='red', label='Train Prec.')
        axs[1][0].plot(training_log_df['batch_total'], training_log_df['recall'], color='blue', label='Train Recall.')
        axs[1][0].plot(training_log_df['batch_total'], training_log_df['r@p99'], color='green', label='Train r@p.')
        axs[1][1].plot(training_log_df['batch_total'], training_log_df['binary_accuracy'], color='black', label='Train Acc.')
        axs[1][1].plot(training_log_df['batch_total'], training_log_df['roc_auc'], color='blue', label='Train ROC AUC')
        axs[1][1].plot(training_log_df['batch_total'], training_log_df['pr_auc'], color='green',
                       label='Train PR AUC')
        if with_validation:
            axs[1][0].plot(validation_log_df['batch_total'], validation_log_df['val_precision'],
                           color='red', label='Val. prec.', linestyle='--')
            axs[1][0].plot(validation_log_df['batch_total'], validation_log_df['val_recall'],
                           color='blue', label='Val. recall.', linestyle='--')
            axs[1][0].plot(validation_log_df['batch_total'], validation_log_df['val_r@p99'],
                           color='green', label='Val. r@p', linestyle='--')
            axs[1][1].plot(validation_log_df['batch_total'], validation_log_df['val_binary_accuracy'],
                           color='black', label='Val. acc.', linestyle='--')
            axs[1][1].plot(validation_log_df['batch_total'], validation_log_df['val_roc_auc'],
                           color='blue', label='Val. ROC AUC', linestyle='--')
            axs[1][1].plot(validation_log_df['batch_total'], validation_log_df['val_pr_auc'],
                           color='green', label='Val. PR AUC', linestyle='--')
        axs[0][0].set_xlabel("Batch")
        secx = axs[0][0].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[0][0].set_ylabel("Learning rate")
        axs[0][0].legend(loc='upper right')
        axs[0][1].set_xlabel("Batch")
        secx = axs[0][1].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[0][1].set_ylabel("Loss")
        axs[0][1].legend(loc='upper right')
        axs[1][0].set_xlabel("Batch")
        secx = axs[1][0].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[1][0].set_ylabel("Precision/Recall/R@P")
        axs[1][0].legend(loc='upper left')
        axs[1][1].set_xlabel("Batch")
        secx = axs[1][1].secondary_xaxis('top',
                                         functions=(lambda x: x / steps_per_epoch, lambda x: x / steps_per_epoch))
        secx.set_xlabel('Epoch')
        axs[1][1].set_ylabel("Accuracy/F1/AUCs")
        axs[1][1].legend(loc='upper left')
        plt.savefig(model_dir + '/metrics.png')
        plt.close(fig)
        plt.clf()
        logging.info("Plotted metrics")



import logging
from abc import ABC
import matplotlib.pyplot as plt
import random
from typing import Tuple
import ellc
import numpy as np
from numpy import ndarray
from numpy.random import default_rng
import astropy.units as u
import astropy.constants as ac
import timesynth
from sklearn.preprocessing import MinMaxScaler


class FeaturesInjector(ABC):
    random_number_generator = default_rng()

    def __init__(self, time_series: ndarray):
        super().__init__()
        self.time_series = time_series

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        pass

class RedNoiseInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray, power: tuple[float, float], period: tuple[float, float], number_of_signals: int):
        super().__init__(time_series)
        self.power = power
        self.period = period
        self.number_of_signals = number_of_signals

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_red_noise = random.uniform(0, 100)
        if number_of_red_noise < 50:
            number_of_red_noise = 0
        elif number_of_red_noise < 75:
            number_of_red_noise = 1
        elif number_of_red_noise < 90:
            number_of_red_noise = 2
        else:
            number_of_red_noise = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_red_noise):
            red_noise_freq: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            red_noise_power: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            logging.info(f'Injecting red noise with power={red_noise_power} and freq={red_noise_freq}')
            red_noise = timesynth.noise.RedNoise(std=red_noise_power, tau=red_noise_freq)
            timeseries_rn: ndarray = np.zeros(self.time_series.shape[1])
            for time_index, value in enumerate(self.time_series[0]):
                rn_value = red_noise.sample_next(value, None, None)
                rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
                timeseries_rn[time_index] = rn_value
            flux = flux + timeseries_rn
        return self.time_series[0], self.time_series[1] + flux - 1, flux


class SinusoidalInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray, amplitude: tuple[float, float], period: tuple[float, float], number_of_signals: int):
        super().__init__(time_series)
        self.power = amplitude
        self.period = period
        self.number_of_signals = number_of_signals

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_signals = random.uniform(0, 100)
        if number_of_signals < 50:
            number_of_signals = 0
        elif number_of_signals < 75:
            number_of_signals = 1
        elif number_of_signals < 90:
            number_of_signals = 2
        else:
            number_of_signals = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_signals):
            period: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            amplitude: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            logging.info(f'Injecting sinusoidal signal with power={amplitude} and freq={period}')
            flux = flux + amplitude * np.sin(2 * np.pi * self.time_series[0] / period + random.uniform(0, 2 * np.pi))
        return self.time_series[0], self.time_series[1] + flux - 1, flux

class PseudoPeriodicInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray, amplitude: tuple[float, float], period: tuple[float, float],
                 amplitude_std: tuple[float, float], period_std: tuple[float, float],
                 number_of_signals: int):
        super().__init__(time_series)
        self.power = amplitude
        self.period = period
        self.number_of_signals = number_of_signals
        self.power_std = amplitude_std
        self.period_std = period_std

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_signals = random.uniform(0, 100)
        if number_of_signals < 50:
            number_of_signals = 0
        elif number_of_signals < 75:
            number_of_signals = 1
        elif number_of_signals < 90:
            number_of_signals = 2
        else:
            number_of_signals = int(random.uniform(0, self.number_of_signals))
        flux: ndarray = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_signals):
            period: float = self.random_number_generator.uniform(self.period[0], self.period[1])
            amplitude: float = self.random_number_generator.uniform(self.power[0], self.power[1])
            period_std: float = self.random_number_generator.uniform(self.period_std[0], self.period_std[1])
            amplitude_std: float = self.random_number_generator.uniform(0, amplitude / 10)
            logging.info(f'Injecting pseudo-periodic signal with power={amplitude} and freq={period}')
            pseudo_periodic_signal = timesynth.signals.PseudoPeriodic(amplitude=amplitude, frequency=period, ampSD=amplitude_std, freqSD=0)
            timeseries_pp: ndarray = np.zeros(self.time_series.shape[1])
            for time_index, value in enumerate(self.time_series[0]):
                rn_value = pseudo_periodic_signal.sample_next(value, None, None)
                rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
                timeseries_pp[time_index] = rn_value
            flux = flux + timeseries_pp
        return self.time_series[0], self.time_series[1] + flux - 1, flux


class TransitsInjector(FeaturesInjector):

    def __init__(self, time_series: ndarray, transit_period_avg_std: tuple[float, float], transit_depth_boundaries: tuple[float, float]):
        super().__init__(time_series)
        self.transit_period_avg_std = transit_period_avg_std
        self.transit_depth_boundaries = transit_depth_boundaries

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_planets = random.uniform(0, 100)
        if number_of_planets < 50:
            number_of_planets = 1
        elif number_of_planets < 80:
            number_of_planets = 2
        else:
            number_of_planets = 3
        star_radius = 1
        star_mass = 1
        flux = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_planets):
            period = 0
            while period < 0.5:
                period = random.gauss(self.transit_period_avg_std[0], self.transit_period_avg_std[1])
            t0 = self.time_series[0][0] + period / random.uniform(0, 1)
            if t0 > self.time_series[0][-1]:
                t0 = np.random.uniform(self.time_series[0][0], self.time_series[0][-1])
            std = np.nanstd(self.time_series[1])
            depth = 0
            while depth < std / 3 or depth >= 0.4:
                depth = random.gauss(self.transit_depth_boundaries[0], self.transit_depth_boundaries[1]) * std
            planet_radius = np.sqrt(depth * (star_radius ** 2))
            logging.info(f'Injecting planet with P={period}, D={depth}')
            P1 = period * u.day
            a = np.cbrt((ac.G * star_mass * u.M_sun * P1 ** 2) / (4 * np.pi ** 2)).to(u.au)
            ld = [0.5, 0.5]
            planet_model = ellc.lc(
                t_obs=self.time_series[0],
                radius_1=(star_radius * u.R_sun).to(u.au) / a,  # star radius convert from AU to in units of a
                radius_2=(planet_radius * u.R_sun).to(u.au) / a,
                # convert from Rearth (equatorial) into AU and then into units of a
                sbratio=0,
                incl=90,
                light_3=0,
                t_zero=t0,
                period=period,
                a=None,
                q=planet_radius / star_radius * 0.1,
                f_c=None, f_s=None,
                ldc_1=ld, ldc_2=None,
                gdc_1=None, gdc_2=None,
                didt=None,
                domdt=None,
                rotfac_1=1, rotfac_2=1,
                hf_1=1.5, hf_2=1.5,
                bfac_1=None, bfac_2=None,
                heat_1=None, heat_2=None,
                lambda_1=None, lambda_2=None,
                vsini_1=None, vsini_2=None,
                t_exp=None, n_int=None,
                grid_1='default', grid_2='default',
                ld_1='quad', ld_2=None,
                shape_1='sphere', shape_2='sphere',
                spots_1=None, spots_2=None,
                exact_grav=False, verbose=1)
            flux = flux + planet_model - 1
        return self.time_series[0], self.time_series[1] + flux - 1, flux


class FlaresInjector(FeaturesInjector):

    def __init__(self, time_series: ndarray,
                 number_of_signals: int,
                 period_avg_std: tuple[float, float],
                 amplitude_avg_std: tuple[float, float],
                 length_avg_std: tuple[float, float]):
        super().__init__(time_series)
        self.number_of_signals = number_of_signals
        self.amplitude_avg_std = amplitude_avg_std
        self.period_avg_std = period_avg_std
        self.length_avg_std = length_avg_std

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        if self.number_of_signals == 0:
            number_of_flares = 0
        else:
            number_of_flares = random.uniform(0, 100)
            if number_of_flares < 50:
                number_of_flares = 1
            elif number_of_flares < 80:
                number_of_flares = 2
            else:
                number_of_flares = int(random.uniform(0, self.number_of_signals))
        flux = np.ones(self.time_series.shape[1])
        for index in range(0, number_of_flares):
            logging.info(f'Injecting flare')
            flux, flares = self.inject_pseudo_periodic_flares(
                self.time_series[0], flux
            )
        return self.time_series[0], self.time_series[1] + flux - 1, flux

    def generate_flare_times(self, T_start, T_end):
        """
        Generate pseudo-periodic flare times between T_start and T_end.
        Periods are drawn from a normal distribution around P_med ± P_std.
        """
        flare_times = []
        current_time = T_start
        while current_time < T_end:
            period = 0
            while period < 5:
                period = random.gauss(self.period_avg_std[0], self.period_avg_std[1])
            current_time += period
            if current_time < T_end:
                flare_times.append(current_time)
        return np.array(flare_times)

    def inject_pseudo_periodic_flares(self, time, flux):
        """
        Inject pseudo-periodic flares with varying amplitude and duration.

        Parameters:
        - times: array of time values
        - fluxes: baseline light curve
        - P_med: median flare period (same units as times)
        - P_std: standard deviation of period
        - amplitude_range: (min, max) relative flare amplitudes
        - t_half_range: (min, max) duration to half-maximum of flares

        Returns:
        - new_fluxes: light curve with flares added
        - flare_metadata: list of dictionaries with flare parameters
        """
        T_start, T_end = time[0], time[-1]
        flare_peaks = self.generate_flare_times(T_start, T_end)
        new_flux = flux.copy()
        flare_metadata = []
        for t_peak in flare_peaks:
            amplitude = 0
            while amplitude < 0.01:
                amplitude = random.gauss(self.amplitude_avg_std[0], self.amplitude_avg_std[1])
            length = 0
            while length < 1 / 24 / 60 * 10 or length > 1 / 24 * 3:
                length = random.gauss(self.length_avg_std[0], self.length_avg_std[1])
            t_half = length / 2
            new_flux = self.inject_flare(time, new_flux, t_peak, amplitude, t_half)
            flare_metadata.append({
                't_peak': t_peak,
                'amplitude': amplitude,
                't_half': t_half
            })
        return new_flux, flare_metadata

    def davenport_flare_model(self, t_normalized):
        """
        Empirical flare shape from Davenport et al. (2014).
        Input: normalized time array (t_peak = 0, normalized by half-time)
        Output: relative flux (normalized to flare peak = 1)
        """
        flux = np.zeros_like(t_normalized)
        # Rise phase (t < 0)
        rise = t_normalized < 0
        t_rise = t_normalized[rise]
        flux[rise] = 1 + 1.941 * t_rise - 0.175 * t_rise ** 2 - 2.246 * t_rise ** 3 - 1.125 * t_rise ** 4
        # Decay phase (t >= 0)
        decay = t_normalized >= 0
        t_decay = t_normalized[decay]
        flux[decay] = 0.689 * np.exp(-1.600 * t_decay) + 0.303 * np.exp(-0.278 * t_decay)
        return flux

    def inject_flare(self, time, flux, t_peak, amplitude, t_half, window_factor=5):
        """
        Inject a synthetic flare into a light curve within a local time window.

        Parameters:
        - times: array of time stamps
        - fluxes: array of flux values
        - t_peak: time of flare peak
        - amplitude: relative flare amplitude (e.g., 0.01 for +1%)
        - t_half: half-duration (same units as times)
        - window_factor: number of t_half widths to define the flare window

        Returns:
        - new_fluxes: array with the flare added
        """
        new_fluxes = flux.copy()
        # Define flare injection window
        mask = np.argwhere((time > t_peak - t_half) &
                           (time < t_peak + t_half * window_factor)).flatten()
        t_norm = (time[mask] - t_peak) / t_half
        flare_shape = self.davenport_flare_model(t_norm)
        flare_flux = amplitude * flare_shape
        new_fluxes[mask] += flare_flux
        return new_fluxes

class GapsInjector(FeaturesInjector):
    def __init__(self, time_series: ndarray,
                 gaps_size_boundaries: tuple[float, float], curve_cadence):
        super().__init__(time_series)
        self.gaps_size_boundaries = gaps_size_boundaries
        self.curve_cadence = curve_cadence

    def inject(self) -> tuple[ndarray, ndarray, ndarray]:
        number_of_gaps = random.uniform(0, 100)
        if number_of_gaps < 50:
            number_of_gaps = 1
        elif number_of_gaps < 80:
            number_of_gaps = 2
        else:
            number_of_gaps = 3
        mask_indexes = []
        for _ in np.arange(0, number_of_gaps):
            gap_size = np.random.uniform(self.gaps_size_boundaries[0], self.gaps_size_boundaries[1])
            gap_size_cadences = int(gap_size / self.curve_cadence)
            gap_start_position = int(np.random.uniform(0, self.time_series.shape[1]))
            new_gap_indexes = np.arange(gap_start_position, min(gap_start_position + gap_size_cadences, self.time_series.shape[1]))
            mask_indexes = list(set(mask_indexes).union(new_gap_indexes))
        mask = np.ones(self.time_series.shape[1], dtype=bool)
        mask[mask_indexes] = False
        flux = self.time_series[1]
        flux_model = self.time_series[2]
        flux[~mask] = 0
        flux_model[~mask] = 0
        return self.time_series[0], flux, flux_model


class SantoTestDataGenerator:
    random_number_generator = default_rng()

    def __init__(self, output_dir: str, minimum_curve_size: int,
                 curves_time_boundaries: Tuple[float, float],
                 transit_depth_boundaries: Tuple[float, float],
                 transit_period_avg_std: Tuple[float, float],
                 white_noise_power_boundaries: Tuple[float, float],
                 red_noise_frequency_boundaries: Tuple[float, float],
                 red_noise_power_boundaries: Tuple[float, float],
                 sinusoid_frequency_boundaries: Tuple[float, float],
                 sinusoid_power_boundaries: Tuple[float, float],
                 pseudo_periodic_frequency_boundaries: Tuple[float, float],
                 pseudo_periodic_power_boundaries: Tuple[float, float],
                 pseudo_periodic_frequency_std_boundaries: Tuple[float, float],
                 pseudo_periodic_power_std_boundaries: Tuple[float, float],
                 gaps_size_boundaries: Tuple[float, float],
                 max_number_of_flares: int = 0,
                 max_number_of_pseudo_periodic: int = 3,
                 max_number_of_sinusoid: int = 3,
                 max_number_of_red_noise: int = 5,
                 flares_period_avg_std: Tuple[float, float] = (10, 5),
                 flares_amplitude_avg_std: Tuple[float, float] = (0.05, 0.05),
                 flares_length_avg_std: Tuple[float, float] = (0.05, 0.05),
                 curves_count=10000, curves_cadences: list=[60, 120, 300, 600, 1800],
                 only_plot=False):
        for curve_index in range(0, curves_count):
            curve_cadences = 1
            while curve_cadences < minimum_curve_size:
                curve_timespan: float = random.uniform(curves_time_boundaries[0], curves_time_boundaries[1])
                selected_cadence = random.choice(curves_cadences) / 3600 / 24
                curve_cadences = int(curve_timespan // selected_cadence)
            flux: ndarray = np.ones((36, curve_cadences))
            flux[0] = np.linspace(0, curve_timespan, curve_cadences)
            white_noise_std: float = 0.0
            while white_noise_std < 0.00001:
                white_noise_std = random.gauss(white_noise_power_boundaries[0], white_noise_power_boundaries[1])
            logging.info(f'Injecting white noise with std={white_noise_std}')
            flux[1] = flux[1] + np.random.normal(loc=0, scale=white_noise_std, size=curve_cadences)
            flux[0], flux[1], flux[2] = RedNoiseInjector(flux, red_noise_power_boundaries, red_noise_frequency_boundaries, max_number_of_red_noise).inject()
            #flux[0] = SinusoidalInjector(time, flux[0], sinusoid_power_boundaries, sinusoid_frequency_boundaries, max_number_of_sinusoid).inject()
            flux[0], flux[1], flux[2] = PseudoPeriodicInjector(flux, pseudo_periodic_power_boundaries, pseudo_periodic_frequency_boundaries,
                                             pseudo_periodic_power_std_boundaries, pseudo_periodic_frequency_std_boundaries, max_number_of_pseudo_periodic).inject()
            flux[0], flux[1], flux[2] = FlaresInjector(flux, max_number_of_flares, flares_period_avg_std, flares_amplitude_avg_std, flares_length_avg_std).inject()
            flux[0], flux[1], flux[2] = TransitsInjector(flux, transit_period_avg_std, transit_depth_boundaries).inject()
            flux[0], flux[1], flux[2] = GapsInjector(flux, gaps_size_boundaries, selected_cadence).inject()
            flux[3] = np.copy(flux[2])
            flux[3, np.argwhere((flux[3] <= 1e-7) | (flux[3] > 1 - 1e-7)).flatten()] = float(0)
            flux[3, np.argwhere(flux[3] > 1e-7).flatten()] = float(1)
            flux[4:36] = self.timestamp_encoding(flux[0])
            if only_plot:
                args_empty = np.argwhere(flux[1] > 0).flatten()
                fig, axs = plt.subplots(1, 1, figsize=(10, 5))
                axs.scatter(flux[0][args_empty], flux[1][args_empty], s=2)
                axs.plot(flux[0][args_empty], flux[2][args_empty], c='orange')
                plt.show()
            else:
                np.savetxt(f'{output_dir}/{curve_index}_model.csv', flux, delimiter=',')

    def timestamp_encoding(self, timestamps: np.ndarray, d_model: int = 32) -> np.ndarray:
        """
        Encode real timestamps into a positional encoding-like representation.

        Args:
            timestamps (np.ndarray): Array of shape (500,) with time values.
            d_model (int): Embedding dimension (must be even).

        Returns:
            np.ndarray: Array of shape (500, d_model) with time embeddings.
        """
        timestamps = timestamps.reshape(-1, 1)  # Ensure shape (500, 1)
        div_term = np.exp(np.arange(0, d_model, 2) * (-np.log(10000.0) / d_model))
        angles = timestamps * div_term
        pos_encoding = np.zeros((timestamps.shape[0], d_model))
        pos_encoding[:, 0::2] = (np.sin(angles) + 1) / 2
        pos_encoding[:, 1::2] = (np.cos(angles) + 1) / 2
        pos_encoding[pos_encoding > 1 - 1e-7] = 1 - 1e-7
        pos_encoding[pos_encoding < 0 + 1e-7] = 1 - 1e-7
        return np.transpose(pos_encoding)




from keras.activations import softmax, sigmoid
from keras.losses import CategoricalCrossentropy, BinaryCrossentropy, binary_crossentropy
from keras.metrics import CategoricalAccuracy, BinaryAccuracy
import tensorflow as tf
from keras import backend as K
import numpy as np


def categorical_cross_entropy_with_dca_loss(y_true, y_pred, alpha=1., beta=10., from_logits=False,
                                            activation_function=softmax):
    """
    DCA loss applied to categorical classification
    :param y_true: the real labels values (typically one hot encoded)
    :param y_pred: the predicted probabilities
    :param alpha: the alpha corrector for the categorical cross entropy loss
    :param beta: the beta corrector for the dca (difference between confidence and accuracy)
    :param from_logits: whether use weights or probabilities
    :param activation_function: the activation function to be applied if from_logits=True
    :return: the dca loss
    """
    return dca_loss(y_true, y_pred, CategoricalCrossentropy(from_logits=from_logits), CategoricalAccuracy(), alpha,
                    beta, from_logits, activation_function)


def binary_cross_entropy_with_dca_loss(y_true, y_pred, alpha=1., beta=10., from_logits=False,
                                       activation_function=sigmoid):
    """
    DCA loss applied to binary classification
    :param y_true: the real labels values (typically one hot encoded)
    :param y_pred: the predicted probabilities
    :param alpha: the alpha corrector for the categorical cross entropy loss
    :param beta: the beta corrector for the dca (difference between confidence and accuracy)
    :param from_logits: whether use weights or probabilities
    :param activation_function: the activation function to be applied if from_logits=True
    :return: the dca loss
    """
    return dca_loss(y_true, y_pred, BinaryCrossentropy(from_logits=from_logits), BinaryAccuracy(), alpha, beta,
                    from_logits, activation_function)


def dca_loss(y_true, y_pred, loss, accuracy, alpha=1., beta=10., from_logits=False, activation_function=softmax):
    """
    Loss function applying penalization to loss decrements when accuracy keeps flatten from Liang et al. (2020)
    https://arxiv.org/abs/2009.04057 (https://stackoverflow.com/questions/68755788/keras-version-of-the-combined-cross-entropy-and-calibration-loss)
    :param y_true: the real labels values (typically one hot encoded)
    :param y_pred: the predicted probabilities
    :param alpha: the alpha corrector for the categorical cross entropy loss
    :param beta: the beta corrector for the dca (difference between confidence and accuracy)
    :param from_logits: whether use weights or probabilities
    :param activation_function: the activation function to be applied if from_logits=True
    :return: the computed loss
    """
    loss_value = loss(y_true, y_pred)
    pred_values = y_pred
    if from_logits and activation_function is not None:
        pred_values = softmax(y_true, axis=1)
    predictions = tf.math.argmax(pred_values, axis=1)
    confidences = tf.reduce_max(pred_values, axis=1)
    mean_conf = tf.reduce_mean(confidences)
    accuracy.update_state(pred_values, y_pred)
    acc = accuracy.result().numpy()
    dca = tf.abs(mean_conf - acc)
    loss_value = alpha * loss_value + beta * dca
    return loss_value


def binary_focal_loss(y_true, y_pred, gamma=2., alpha=.2):
    """
    Binary form of focal loss.
      FL(p_t) = -alpha * (1 - p_t)**gamma * log(p_t)
      where p = sigmoid(x), p_t = p or 1 - p depending on if the label is 1 or 0, respectively.
    References:
        https://arxiv.org/pdf/1708.02002.pdf
    Usage:
     model.compile(loss=[binary_focal_loss(alpha=.25, gamma=2)], metrics=["accuracy"], optimizer=adam)
    """
    y_true = tf.cast(y_true, tf.float32)
    # Define epsilon so that the back-propagation will not result in NaN for 0 divisor case
    epsilon = K.epsilon()
    # Add the epsilon to prediction value
    # y_pred = y_pred + epsilon
    # Clip the prediciton value
    y_pred = K.clip(y_pred, epsilon, 1.0 - epsilon)
    # Calculate p_t
    p_t = tf.where(K.equal(y_true, 1), y_pred, 1 - y_pred)
    # Calculate alpha_t
    alpha_factor = K.ones_like(y_true) * alpha
    alpha_t = tf.where(K.equal(y_true, 1), alpha_factor, 1 - alpha_factor)
    # Calculate cross entropy
    cross_entropy = -K.log(p_t)
    weight = alpha_t * K.pow((1 - p_t), gamma)
    # Calculate focal loss
    loss = weight * cross_entropy
    # Sum the losses in mini_batch
    loss = K.mean(K.sum(loss, axis=1))
    return loss


def categorical_focal_loss(y_true, y_pred, alpha=0.2, gamma=2.):
    """
    Softmax version of focal loss.
    When there is a skew between different categories/labels in your data set, you can try to apply this function as a
    loss.
           m
      FL = ∑  -alpha * (1 - p_o,c)^gamma * y_o,c * log(p_o,c)
          c=1
      where m = number of classes, c = class and o = observation
    Parameters:
      alpha -- the same as weighing factor in balanced cross entropy. Alpha is used to specify the weight of different
      categories/labels, the size of the array needs to be consistent with the number of classes.
      gamma -- focusing parameter for modulating factor (1-p)
    Default value:
      gamma -- 2.0 as mentioned in the paper
      alpha -- 0.25 as mentioned in the paper
    References:
        Official paper: https://arxiv.org/pdf/1708.02002.pdf
        https://www.tensorflow.org/api_docs/python/tf/keras/backend/categorical_crossentropy
    Usage:
     model.compile(loss=[categorical_focal_loss(alpha=[[.25, .25, .25]], gamma=2)], metrics=["accuracy"], optimizer=adam)
    """
    # Clip the prediction value to prevent NaN's and Inf's
    epsilon = K.epsilon()
    y_pred = K.clip(y_pred, epsilon, 1. - epsilon)
    # Calculate Cross Entropy
    cross_entropy = -y_true * K.log(y_pred)
    # Calculate Focal Loss
    loss = alpha * K.pow(1 - y_pred, gamma) * cross_entropy
    # Compute mean loss in mini_batch
    return K.mean(K.sum(loss, axis=-1))


def fb_loss(y_true, y_pred, beta=1):
    """
    The macro F1-score has one big trouble. It's non-differentiable. Which means we cannot use it as a loss function.

    But we can modify it to be differentiable. Instead of accepting 0/1 integer predictions, let's accept probabilities
    instead. Thus if the ground truth is 1 and the model prediction is 0.4, we calculate it as 0.4 true positive and
    0.6 false negative. If the ground truth is 0 and the model prediction is 0.4, we calculate it as 0.6 true negative
    and 0.4 false positive.
    :param y_true: the real outputs
    :param y_pred: the predicted outputs
    :return:
    """
    tp = K.sum(K.cast(y_true * y_pred, 'float'), axis=0)
    tn = K.sum(K.cast((1 - y_true) * (1 - y_pred), 'float'), axis=0)
    fp = K.sum(K.cast((1 - y_true) * y_pred, 'float'), axis=0)
    fn = K.sum(K.cast(y_true * (1 - y_pred), 'float'), axis=0)
    p = tp / (tp + fp + K.epsilon())
    r = tp / (tp + fn + K.epsilon())
    f1 = (1 + beta ** 2) * p * r / (beta ** 2 * p + r + K.epsilon())
    f1 = tf.where(tf.math.is_nan(f1), tf.zeros_like(f1), f1)
    return 1 - K.mean(f1)


def dice_loss(targets, inputs, smooth=1e-6):
    """
    From https://www.kaggle.com/code/bigironsphere/loss-function-library-keras-pytorch/notebook
    :param targets: the expected outputs
    :param inputs: the given inputs
    :param smooth:
    :return: the dice loss
    """
    inputs = K.flatten(inputs)
    targets = K.flatten(targets)
    intersection = K.sum((inputs * targets))
    dice = (2 * intersection + smooth) / (K.sum(targets) + K.sum(inputs) + smooth)
    return 1 - dice


def dice_bce_loss(targets, inputs, smooth=1e-6):
    """
    This loss combines Dice loss with the standard binary cross-entropy (BCE) loss that is generally the default for
    segmentation models. Combining the two methods allows for some diversity in the loss, while benefitting from the
    stability of BCE.
    From https://www.kaggle.com/code/bigironsphere/loss-function-library-keras-pytorch/notebook
    :param targets: the expected outputs
    :param inputs: the given inputs
    :param smooth: the dice loss smooth parameter
    :return:
    """
    inputs = K.flatten(inputs)
    targets = K.flatten(targets)
    bce = binary_crossentropy(targets, inputs)
    intersection = K.sum((inputs * targets))
    dice_loss = 1 - (2 * intersection + smooth) / (K.sum(targets) + K.sum(inputs) + smooth)
    dice_bce = bce + dice_loss
    return dice_bce





import abc
import importlib
import warnings

#from keras import backend
#from keras_core.src.utils import tf_utils
from keras.src import backend
from keras.src.utils import tf_utils

from keras.callbacks import Callback
import tensorflow as tf
#from tensorflow_addons.optimizers import MultiOptimizer
from typeguard import typechecked
from typing import Union, List


if (
    hasattr(tf.keras.optimizers, "experimental")
    and tf.keras.optimizers.Optimizer.__module__
    == tf.keras.optimizers.experimental.Optimizer.__module__
):
    # If the default optimizer points to new Keras optimizer, addon optimizers
    # should use the legacy path.
    KerasLegacyOptimizer = tf.keras.optimizers.legacy.Optimizer
else:
    KerasLegacyOptimizer = tf.keras.optimizers.Optimizer


if importlib.util.find_spec("tensorflow.keras.optimizers.legacy") is not None:
    Optimizer = Union[
        tf.keras.optimizers.Optimizer, tf.keras.optimizers.legacy.Optimizer, str
    ]
else:
    Optimizer = Union[tf.keras.optimizers.Optimizer, str]


class AveragedOptimizerWrapper(KerasLegacyOptimizer, metaclass=abc.ABCMeta):
    @typechecked
    def __init__(
        self,
        optimizer: Optimizer,
        name: str = "AverageOptimizer",
        **kwargs,
    ):
        super().__init__(name, **kwargs)

        if isinstance(optimizer, str):
            if (
                hasattr(tf.keras.optimizers, "legacy")
                and KerasLegacyOptimizer == tf.keras.optimizers.legacy.Optimizer
            ):
                optimizer = tf.keras.optimizers.get(
                    optimizer, use_legacy_optimizer=True
                )
            else:
                optimizer = tf.keras.optimizers.get(optimizer)

        if not isinstance(optimizer, KerasLegacyOptimizer):
            raise TypeError(
                "optimizer is not an object of tf.keras.optimizers.legacy.Optimizer "
            )

        self._optimizer = optimizer
        self._track_trackable(self._optimizer, "awg_optimizer")

    def _create_slots(self, var_list):
        self._optimizer._create_slots(var_list=var_list)
        for var in var_list:
            self.add_slot(var, "average")

    def _create_hypers(self):
        self._optimizer._create_hypers()

    def _prepare_local(self, var_device, var_dtype, apply_state):
        return self._optimizer._prepare_local(var_device, var_dtype, apply_state)

    def apply_gradients(self, grads_and_vars, name=None, **kwargs):
        self._optimizer._iterations = self.iterations
        return super().apply_gradients(grads_and_vars, name, **kwargs)

    @abc.abstractmethod
    def average_op(self, var, average_var, local_apply_state):
        raise NotImplementedError

    def _apply_average_op(self, train_op, var, apply_state):
        apply_state = apply_state or {}
        local_apply_state = apply_state.get((var.device, var.dtype.base_dtype))
        if local_apply_state is None:
            local_apply_state = self._fallback_apply_state(
                var.device, var.dtype.base_dtype
            )
        average_var = self.get_slot(var, "average")
        return self.average_op(var, average_var, local_apply_state)

    def _resource_apply_dense(self, grad, var, apply_state=None):
        if "apply_state" in self._optimizer._dense_apply_args:
            train_op = self._optimizer._resource_apply_dense(
                grad, var, apply_state=apply_state
            )
        else:
            train_op = self._optimizer._resource_apply_dense(grad, var)
        average_op = self._apply_average_op(train_op, var, apply_state)
        return tf.group(train_op, average_op)

    def _resource_apply_sparse(self, grad, var, indices, apply_state=None):
        if "apply_state" in self._optimizer._sparse_apply_args:
            train_op = self._optimizer._resource_apply_sparse(
                grad, var, indices, apply_state=apply_state
            )
        else:
            train_op = self._optimizer._resource_apply_sparse(grad, var, indices)
        average_op = self._apply_average_op(train_op, var, apply_state)
        return tf.group(train_op, average_op)

    def _resource_apply_sparse_duplicate_indices(
        self, grad, var, indices, apply_state=None
    ):
        if "apply_state" in self._optimizer._sparse_apply_args:
            train_op = self._optimizer._resource_apply_sparse_duplicate_indices(
                grad, var, indices, apply_state=apply_state
            )
        else:
            train_op = self._optimizer._resource_apply_sparse_duplicate_indices(
                grad, var, indices
            )
        average_op = self._apply_average_op(train_op, var, apply_state)
        return tf.group(train_op, average_op)

    def assign_average_vars(self, var_list):
        """Assign variables in var_list with their respective averages.

        Args:
            var_list: List of model variables to be assigned to their average.

        Returns:
            assign_op: The op corresponding to the assignment operation of
            variables to their average.

        Example:
        ```python
        model = tf.Sequential([...])
        opt = tfa.optimizers.SWA(
                tf.keras.optimizers.SGD(lr=2.0), 100, 10)
        model.compile(opt, ...)
        model.fit(x, y, ...)

        # Update the weights to their mean before saving
        opt.assign_average_vars(model.variables)

        model.save('model.h5')
        ```
        """
        assign_ops = []
        for var in var_list:
            try:
                assign_ops.append(
                    var.assign(
                        self.get_slot(var, "average"),
                        use_locking=self._use_locking,
                    )
                )
            except Exception as e:
                warnings.warn("Unable to assign average slot to {} : {}".format(var, e))
        return tf.group(assign_ops)

    def get_config(self):
        config = {
            "optimizer": tf.keras.optimizers.serialize(self._optimizer),
        }
        base_config = super().get_config()
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config, custom_objects=None):
        optimizer = tf.keras.optimizers.deserialize(
            config.pop("optimizer"), custom_objects=custom_objects
        )
        return cls(optimizer, **config)

    @property
    def weights(self):
        return self._weights + self._optimizer.weights

    @property
    def lr(self):
        return self._optimizer._get_hyper("learning_rate")

    @lr.setter
    def lr(self, lr):
        self._optimizer._set_hyper("learning_rate", lr)  #

    @property
    def learning_rate(self):
        return self._optimizer._get_hyper("learning_rate")

    @learning_rate.setter
    def learning_rate(self, learning_rate):
        self._optimizer._set_hyper("learning_rate", learning_rate)


@tf.keras.utils.register_keras_serializable(package="Addons")
class SWA(AveragedOptimizerWrapper):
     """This class extends optimizers with Stochastic Weight Averaging (SWA).

     The Stochastic Weight Averaging mechanism was proposed by Pavel Izmailov
     et. al in the paper [Averaging Weights Leads to Wider Optima and
     Better Generalization](https://arxiv.org/abs/1803.05407). The optimizer
     implements averaging of multiple points along the trajectory of SGD. The
    optimizer expects an inner optimizer which will be used to apply the
     gradients to the variables and itself computes a running average of the
     variables every `k` steps (which generally corresponds to the end
     of a cycle when a cyclic learning rate is employed).

     We also allow the specification of the number of steps averaging
     should first happen after. Let's say, we want averaging to happen every `k`
     steps after the first `m` steps. After step `m` we'd take a snapshot of the
     variables and then average the weights appropriately at step `m + k`,
     `m + 2k` and so on. The assign_average_vars function can be called at the
     end of training to obtain the averaged_weights from the optimizer.

     Note: If your model has batch-normalization layers you would need to run
     the final weights through the data to compute the running mean and
     variance corresponding to the activations for each layer of the network.
     From the paper: If the DNN uses batch normalization we run one
     additional pass over the data, to compute the running mean and standard
     deviation of the activations for each layer of the network with SWA
     weights after the training is finished, since these statistics are not
     collected during training. For most deep learning libraries, such as
     PyTorch or Tensorflow, one can typically collect these statistics by
     making a forward pass over the data in training mode
     ([Averaging Weights Leads to Wider Optima and Better
     Generalization](https://arxiv.org/abs/1803.05407))

     Example of usage:

     ```python
     opt = tf.keras.optimizers.SGD(learning_rate)
     opt = tfa.optimizers.SWA(opt, start_averaging=m, average_period=k)
     ```
     """

     @typechecked
     def __init__(
         self,
         optimizer: Optimizer,
         start_averaging: int = 0,
         average_period: int = 10,
         name: str = "SWA",
         **kwargs,
     ):
         r"""Wrap optimizer with the Stochastic Weight Averaging mechanism.

         Args:
             optimizer: The original optimizer that will be used to compute and
                 apply the gradients.
             start_averaging: An integer. Threshold to start averaging using
                 SWA. Averaging only occurs at `start_averaging` iters, must
                 be >= 0. If start_averaging = m, the first snapshot will be
                 taken after the mth application of gradients (where the first
                 iteration is iteration 0).
             average_period: An integer. The synchronization period of SWA. The
                 averaging occurs every average_period steps. Averaging period
                 needs to be >= 1.
             name: Optional name for the operations created when applying
                 gradients. Defaults to 'SWA'.
             **kwargs: keyword arguments. Allowed to be {`clipnorm`,
                 `clipvalue`, `lr`, `decay`}. `clipnorm` is clip gradients by
                 norm; `clipvalue` is clip gradients by value, `decay` is
                 included for backward compatibility to allow time inverse
                 decay of learning rate. `lr` is included for backward
                 compatibility, recommended to use `learning_rate` instead.
         """
         super().__init__(optimizer, name, **kwargs)

         if average_period < 1:
             raise ValueError("average_period must be >= 1")
         if start_averaging < 0:
             raise ValueError("start_averaging must be >= 0")

         self._set_hyper("average_period", average_period)
         self._set_hyper("start_averaging", start_averaging)

     @tf.function
     def average_op(self, var, average_var, local_apply_state):
         average_period = self._get_hyper("average_period", tf.dtypes.int64)
         start_averaging = self._get_hyper("start_averaging", tf.dtypes.int64)
         # number of times snapshots of weights have been taken (using max to
         # avoid negative values of num_snapshots).
         num_snapshots = tf.math.maximum(
             tf.cast(0, tf.int64),
             tf.math.floordiv(self.iterations - start_averaging, average_period),
         )

         # The average update should happen iff two conditions are met:
         # 1. A min number of iterations (start_averaging) have taken place.
         # 2. Iteration is one in which snapshot should be taken.
         checkpoint = start_averaging + num_snapshots * average_period
         if self.iterations >= start_averaging and self.iterations == checkpoint:
             num_snapshots = tf.cast(num_snapshots, tf.float32)
             average_value = (average_var * num_snapshots + var) / (num_snapshots + 1.0)
             return average_var.assign(average_value, use_locking=self._use_locking)

         return average_var

     def get_config(self):
         config = {
             "average_period": self._serialize_hyperparameter("average_period"),
             "start_averaging": self._serialize_hyperparameter("start_averaging"),
         }
         base_config = super().get_config()
         return {**base_config, **config}

@tf.keras.utils.register_keras_serializable(package="Addons")
class MultiOptimizer(KerasLegacyOptimizer):
     """Multi Optimizer Wrapper for Discriminative Layer Training.

     Creates a wrapper around a set of instantiated optimizer layer pairs.
     Generally useful for transfer learning of deep networks.

     Each optimizer will optimize only the weights associated with its paired layer.
     This can be used to implement discriminative layer training by assigning
     different learning rates to each optimizer layer pair.
     `(tf.keras.optimizers.legacy.Optimizer, List[tf.keras.layers.Layer])` pairs are also supported.
     Please note that the layers must be instantiated before instantiating the optimizer.

     Args:
         optimizers_and_layers: a list of tuples of an optimizer and a layer or model.
             Each tuple should contain exactly 1 instantiated optimizer and 1 object that
             subclasses `tf.keras.Model`, `tf.keras.Sequential` or `tf.keras.layers.Layer`.
             Nested layers and models will be automatically discovered.
             Alternatively, in place of a single layer, you can pass a list of layers.
         optimizer_specs: specialized list for serialization.
             Should be left as None for almost all cases.
             If you are loading a serialized version of this optimizer,
             please use `tf.keras.models.load_model` after saving a model compiled with this optimizer.

     Usage:

     >>> model = tf.keras.Sequential([
     ...     tf.keras.Input(shape=(4,)),
     ...     tf.keras.layers.Dense(8),
     ...     tf.keras.layers.Dense(16),
     ...     tf.keras.layers.Dense(32),
     ... ])
     >>> optimizers = [
     ...     tf.keras.optimizers.Adam(learning_rate=1e-4),
     ...     tf.keras.optimizers.Adam(learning_rate=1e-2)
     ... ]
     >>> optimizers_and_layers = [(optimizers[0], model.layers[0]), (optimizers[1], model.layers[1:])]
     >>> optimizer = tfa.optimizers.MultiOptimizer(optimizers_and_layers)
     >>> model.compile(optimizer=optimizer, loss="mse")

     Reference:
         - [Universal Language Model Fine-tuning for Text Classification](https://arxiv.org/abs/1801.06146)
         - [Collaborative Layer-wise Discriminative Learning in Deep Neural Networks](https://arxiv.org/abs/1607.05440)

     Note: Currently, `tfa.optimizers.MultiOptimizer` does not support callbacks that modify optimizers.
         However, you can instantiate optimizer layer pairs with
         `tf.keras.optimizers.schedules.LearningRateSchedule`
         instead of a static learning rate.

     This code should function on CPU, GPU, and TPU. Apply with `tf.distribute.Strategy().scope()` context as you
     would with any other optimizer.
     """

     @typechecked
     def __init__(
          self,
         optimizers_and_layers: Union[list, None] = None,
         optimizer_specs: Union[list, None] = None,
         name: str = "MultiOptimizer",
         **kwargs,
     ):

         super(MultiOptimizer, self).__init__(name, **kwargs)

         if optimizer_specs is None and optimizers_and_layers is not None:
             self.optimizer_specs = [
                 self.create_optimizer_spec(optimizer, layers_or_model)
                 for optimizer, layers_or_model in optimizers_and_layers
             ]

         elif optimizer_specs is not None and optimizers_and_layers is None:
             self.optimizer_specs = [
                 self.maybe_initialize_optimizer_spec(spec) for spec in optimizer_specs
             ]

         else:
             raise RuntimeError(
                 "Must specify one of `optimizers_and_layers` or `optimizer_specs`."
             )

     def apply_gradients(self, grads_and_vars, **kwargs):
         """Wrapped apply_gradient method.

         Returns an operation to be executed.
         """

         for spec in self.optimizer_specs:
             spec["gv"] = []

         for grad, var in tuple(grads_and_vars):
             for spec in self.optimizer_specs:
                 for name in spec["weights"]:
                     if var.name == name:
                         spec["gv"].append((grad, var))

         update_ops = [
             spec["optimizer"].apply_gradients(spec["gv"], **kwargs)
             for spec in self.optimizer_specs
         ]
         update_group = tf.group(update_ops)

         any_symbolic = any(
             isinstance(i, tf.Operation) or tf_utils.is_symbolic_tensor(i)
             for i in update_ops
         )

         if not tf.executing_eagerly() or any_symbolic:
             # If the current context is graph mode or any of the update ops are
             # symbolic then the step update should be carried out under a graph
             # context. (eager updates execute immediately)
             with backend._current_graph(  # pylint: disable=protected-access
                 update_ops
             ).as_default():
                 with tf.control_dependencies([update_group]):
                     return self.iterations.assign_add(1, read_value=False)

         return self.iterations.assign_add(1)

     def get_config(self):
         config = super(MultiOptimizer, self).get_config()
         optimizer_specs_without_gv = []
         for optimizer_spec in self.optimizer_specs:
             optimizer_specs_without_gv.append(
                 {
                     "optimizer": optimizer_spec["optimizer"],
                     "weights": optimizer_spec["weights"],
                 }
             )
         config.update({"optimizer_specs": optimizer_specs_without_gv})
         return config

     @classmethod
     def create_optimizer_spec(
         cls,
         optimizer: KerasLegacyOptimizer,
         layers_or_model: Union[
             tf.keras.Model,
             tf.keras.Sequential,
             tf.keras.layers.Layer,
             List[tf.keras.layers.Layer],
         ],
     ):
         """Creates a serializable optimizer spec.

         The name of each variable is used rather than `var.ref()` to enable serialization and deserialization.
         """
         if isinstance(layers_or_model, list):
             weights = [
                 var.name for sublayer in layers_or_model for var in sublayer.weights
             ]
         else:
             weights = [var.name for var in layers_or_model.weights]

         return {
             "optimizer": optimizer,
             "weights": weights,
         }

     @classmethod
     def maybe_initialize_optimizer_spec(cls, optimizer_spec):
         if isinstance(optimizer_spec["optimizer"], dict):
             optimizer_spec["optimizer"] = tf.keras.optimizers.deserialize(
                 optimizer_spec["optimizer"]
             )

         return optimizer_spec

     def __repr__(self):
         return "Multi Optimizer with %i optimizer layer pairs" % len(
             self.optimizer_specs
         )

class WarmUpAndLinDecreaseCallback(Callback):
    def __init__(
            self,
            initial_lr,
            top_lr,
            baseline_lr,
            warmup_epochs,
            baseline_epochs,
            name=None):
        """Applies linear decrease with a first stage of warmup
    """
        super(WarmUpAndLinDecreaseCallback, self).__init__()
        assert initial_lr > 0
        self.initial_lr = initial_lr
        self.top_lr = top_lr
        self.baseline_lr = baseline_lr
        self.warmup_epochs = warmup_epochs
        self.baseline_epochs = baseline_epochs
        self.name = name
        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        if self.model.optimizer.__class__.__name__ == MultiOptimizer.__name__:
            for index, optimizer_spec in enumerate(self.model.optimizer.optimizer_specs):
                optimizer = optimizer_spec['optimizer']
                warmup_slope = (self.top_lr - self.initial_lr) / self.warmup_epochs
                decreasing_slope = (self.baseline_lr - self.top_lr) / self.baseline_epochs
                new_lr_decreasing = epoch * decreasing_slope + self.top_lr
                new_lr_warmup = epoch * warmup_slope + self.initial_lr
                new_learning_rate = new_lr_warmup if epoch <= self.warmup_epochs else new_lr_decreasing
                new_learning_rate = new_learning_rate if new_learning_rate < self.baseline_lr else new_learning_rate
                new_learning_rate = new_learning_rate * optimizer.progressive_lr_factor
                backend.set_value(optimizer.lr, new_learning_rate)
        else:
            optimizer = self.model.optimizer
            if self.warmup_epochs <= 0:
                warmup_slope = 1
            else:
                warmup_slope = (self.top_lr - self.initial_lr) / self.warmup_epochs
            decreasing_slope = (self.baseline_lr - self.top_lr) / self.baseline_epochs
            new_lr_decreasing = epoch * decreasing_slope + self.top_lr
            new_lr_warmup = epoch * warmup_slope + self.initial_lr
            new_learning_rate = new_lr_warmup if epoch <= self.warmup_epochs else new_lr_decreasing
            new_learning_rate = new_learning_rate if new_learning_rate < self.baseline_lr else new_learning_rate
            backend.set_value(optimizer.lr, new_learning_rate)



from keras.metrics import CategoricalAccuracy
from exoml.ml.loss.cross_entropy import categorical_focal_loss
from exoml.ml.model.categorical_model import CategoricalModel


class ImbalancedCategoricalModel(CategoricalModel):
    def __init__(self, class_names, name, input_size, class_ids, dropout_rate=0.1) -> None:
        super().__init__(class_names, name, input_size, class_ids, dropout_rate)

    def instance_loss_accuracy(self):
        return categorical_focal_loss, CategoricalAccuracy()



import logging
from abc import abstractmethod

import foldedleastsquares
import scipy
#import timesynth
from lcbuilder.helper import LcbuilderHelper
from numpy.random import default_rng
from scipy import stats
from sklearn.preprocessing import StandardScaler, MinMaxScaler

from exoml.ete6.NanException import NanException
from sklearn.utils import shuffle
import tensorflow as tf
import numpy as np
import pandas as pd

from exoml.ml.functions.functions import sigmoid


class Ete6ModelGenerator(tf.keras.utils.Sequence):
    random_number_generator = default_rng()
    standard_scaler = StandardScaler()

    def __init__(self, zero_epsilon=1e-6, shuffle_batch=True, flag_incomplete=False, throw_nan_exception=False):
        self.zero_epsilon = zero_epsilon
        self.minmax_scaler = MinMaxScaler(feature_range=(self.zero_epsilon, 1 - self.zero_epsilon))
        self.shuffle_batch = shuffle_batch
        self.flag_incomplete = flag_incomplete
        self.throw_nan_exception = throw_nan_exception

    def on_epoch_end(self):
        if self.shuffle_batch:
            self.injected_objects_df = shuffle(self.injected_objects_df)

    def assert_in_range(self, object_id, array, array_err, values_range=(0, 1)):
        if np.isnan(array).any() or (array_err is not None and np.isnan(array_err).any()):
            raise ValueError("Target " + str(object_id) + " contains NaN values")
        elif np.max(array) > values_range[1] or (array_err is not None and np.max(array_err) > values_range[1]):
            raise ValueError("Target " + str(object_id) + " contains values > 1")
        elif np.min(array) < values_range[0] or (array_err is not None and np.max(array_err) < values_range[0]):
            raise ValueError("Target " + str(object_id) + " contains values < 0")
        elif np.all(array == values_range[0]) or (array_err is not None and np.all(array_err == values_range[0])):
            raise ValueError("Target " + str(object_id) + " contains all values == 0")


    def fold(self, time, period, epoch):
        return foldedleastsquares.fold(time, period, epoch) if len(time) > 0 else time

    def bin_interp(self, array, new_size):
        result = np.zeros((array.shape[1], new_size))
        for index, nested_array in enumerate(np.transpose(array)):
            try:
                nested_array = nested_array[nested_array > self.zero_epsilon]
                arr_interp = scipy.interpolate.interp1d(np.arange(nested_array.size), nested_array)
                result[index] = arr_interp(np.linspace(0, nested_array.size - 1, new_size))
            except Exception as e:
                a = 1
                # logging.exception("Error when interpolating, will create empty input array")
        return np.transpose(result)

    def bin_by_time(self, array, new_size, target, min_points=50):
        result = np.full((array.shape[1], new_size), self.zero_epsilon)
        result_err = np.full((array.shape[1], new_size), self.zero_epsilon)
        transposed_array = np.transpose(array)
        time_array = transposed_array[0]
        any_nan = False
        for index, nested_array in enumerate(transposed_array):
            if index > 0:
                mask = nested_array > self.zero_epsilon
                nested_array = nested_array[mask]
                masked_time_array = time_array[mask]
                if len(masked_time_array) < min_points:
                    continue
                try:
                    bin_means, bin_edges, binnumber = stats.binned_statistic(masked_time_array, nested_array,
                                                                             statistic='mean', bins=new_size)
                    bin_stds, _, _ = stats.binned_statistic(masked_time_array, nested_array, statistic='std',
                                                            bins=new_size)
                    nans, x = self.nan_helper(bin_means)
                    any_nan = any_nan or np.any(nans)
                    bin_means[nans] = np.interp(x(nans), x(~nans), bin_means[~nans])
                    nans, x = self.nan_helper(bin_stds)
                    any_nan = any_nan or np.any(nans)
                    bin_stds[nans] = np.interp(x(nans), x(~nans), bin_stds[~nans])
                    bin_width = (bin_edges[1] - bin_edges[0])
                    bin_centers = bin_edges[1:] - bin_width / 2
                    if np.max(result[0]) == self.zero_epsilon:
                        result[0] = bin_centers
                        result_err[0] = np.full((new_size), 0.5)
                    # Store binning values for flux
                    result[index] = self.minmax_scaler.fit_transform(bin_means.reshape(-1, 1)).flatten()
                    result_err[index] = bin_stds if len(np.isnan(bin_stds)) == 0 else np.full((new_size), np.nanstd(bin_means))
                except Exception as e:
                    logging.exception("Error when interpolating for target " + str(target) + ", will create empty input array")
        if self.throw_nan_exception and any_nan:
            raise NanException(f"Nan values in array for target {target}")
        return np.transpose(np.nan_to_num(result[1:])), np.transpose(np.nan_to_num(result_err[1:]))

    @staticmethod
    def nan_helper(y):
        """Helper to handle indices and logical indices of NaNs.

        Input:
            - y, 1d numpy array with possible NaNs
        Output:
            - nans, logical indices of NaNs
            - index, a function, with signature indices= index(logical_indices),
              to convert logical indices of NaNs to 'equivalent' indices
        Example:
            >>> # linear interpolation of NaNs
            >>> nans, x= nan_helper(y)
            >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
        """
        return np.isnan(y), lambda z: z.nonzero()[0]

    def _prepare_input_centroids(self, input_df):
        input_df_clipped = input_df
        if len(input_df) > 0:
            _, ra_outliers_mask = LcbuilderHelper.clip_outliers(input_df['centroids_ra'].to_numpy(), 3)
            _, dec_outliers_mask = LcbuilderHelper.clip_outliers(input_df['centroids_dec'].to_numpy(), 3)
            outliers_mask = np.logical_or(ra_outliers_mask, dec_outliers_mask)
            input_df_clipped = input_df.loc[~outliers_mask].copy()
            if len(input_df_clipped) > 0:
                # transform data
                input_df_clipped['centroids_ra'] = self.minmax_scaler.fit_transform(input_df_clipped['centroids_ra'].values.reshape(-1, 1)).flatten()
                input_df_clipped['centroids_dec'] = self.minmax_scaler.fit_transform(input_df_clipped['centroids_dec'].values.reshape(-1, 1)).flatten()
                input_df_clipped = input_df_clipped.fillna(self.zero_epsilon)
                input_df_clipped = input_df_clipped.replace(0.0, self.zero_epsilon)
                input_df_clipped = input_df_clipped.replace(0, self.zero_epsilon)
        return input_df_clipped

    def _prepare_input_og(self, input_df):
        input_df_clipped = input_df
        if len(input_df) > 0:
            _, outliers_mask = LcbuilderHelper.clip_outliers(input_df['og_flux'].to_numpy(), 3)
            input_df_clipped = input_df.loc[~outliers_mask].copy()
            if len(input_df_clipped) > 0:
                input_df_clipped['og_flux'] = self.minmax_scaler.fit_transform(input_df_clipped['og_flux'].values.reshape(-1, 1)).flatten()
            input_df_clipped = input_df_clipped.fillna(self.zero_epsilon)
            input_df_clipped = input_df_clipped.replace(0.0, self.zero_epsilon)
            input_df_clipped = input_df_clipped.replace(0, self.zero_epsilon)
        return input_df_clipped

    def _prepare_input_lc(self, input_df, period, epoch, duration, focus_range_in_durations=6,
                          required_range_percent=0.75, time_key='#time', flux_key='flux_0'):
        time = input_df[time_key].to_numpy()
        cadence_s = np.round(np.nanmedian(time[1:] - time[:-1]) * 3600 * 24)
        cadences_per_transit = LcbuilderHelper.estimate_transit_cadences(cadence_s, duration * focus_range_in_durations)
        transit_t0s_list = LcbuilderHelper.compute_t0s(time, period, epoch, duration)
        t0s_in_curve_indexes = np.argwhere((transit_t0s_list > time[0] - duration) &
                                          (transit_t0s_list < time[-1] + duration)).flatten()
        transit_t0s_list = transit_t0s_list[t0s_in_curve_indexes]
        good_quality_t0s = []
        t0s_with_data = []
        for t0 in transit_t0s_list:
            transit_with_baseline_half_length = (duration * focus_range_in_durations / 2)
            t0_in_curve_indexes = np.argwhere((time > t0 - transit_with_baseline_half_length) &
                                              (time < t0 + transit_with_baseline_half_length)).flatten()
            cadences_ratio = len(t0_in_curve_indexes) / cadences_per_transit
            if cadences_ratio > 0:
                t0s_with_data.append(t0)
            if cadences_ratio >= required_range_percent:
                good_quality_t0s.append(t0)
            else:
                input_df = input_df.loc[(input_df[time_key] < t0 - transit_with_baseline_half_length) |
                                    (input_df[time_key] > t0 + transit_with_baseline_half_length)]

        # We don't want time encoding because time will be used for folding and then dropped
        # input_df["#time"] = value_encode_times(time - time[0])
        # red_noise_freq = self.random_number_generator.uniform(0.01, 0.1)
        # flux_std = np.std(input_df['flux'])
        # red_noise_sd = self.random_number_generator.uniform(0, flux_std * 3)
        # red_noise = timesynth.noise.RedNoise(std=red_noise_sd, tau=red_noise_freq)
        # timeseries_rn = np.zeros(len(time))
        # TODO red noise algorithm is too slow for long time series
        # timeseries_rn = []
        # for value in time:
        #     rn_value = red_noise.sample_next(value, None, None)
        #     rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
        #     timeseries_rn = timeseries_rn + [rn_value]
        # timeseries_rn = np.array(timeseries_rn)
        # input_df['flux_0'] = input_df['flux_0'] + timeseries_rn
        input_df[flux_key] = input_df[flux_key] / 2
        input_df[flux_key][input_df[flux_key] > 1] = 1
        input_df = input_df.fillna(self.zero_epsilon)
        input_df = input_df.replace(0.0, self.zero_epsilon)
        input_df = input_df.replace(0, self.zero_epsilon)
        return input_df, len(good_quality_t0s), len(t0s_with_data)

    def _prepare_input_star(self, target_row, star_df, teff_mod=1, rad_mod=1, mass_mod=1):
        if 'ra' in star_df.columns:
            ra = star_df['ra'].iloc[0]
        if 'dec' in star_df.columns:
            dec = star_df['dec'].iloc[0]
        if 'id' in star_df.columns:
            star_df.drop('id', inplace=True, axis=1)
        elif 'obj_id' in star_df.columns:
            star_df.drop('obj_id', inplace=True, axis=1)
        if 'Unnamed: 0' in star_df.columns:
            star_df.drop('Unnamed: 0', inplace=True, axis=1)
        if 'ra' in star_df.columns:
            star_df.drop('ra', inplace=True, axis=1)
        if pd.Series(['ra,']).isin(star_df.columns).any():
            star_df.drop('ra,', inplace=True, axis=1)
        if 'dec' in star_df.columns:
            star_df.drop('dec', inplace=True, axis=1)
        if 'dist_arcsec' in star_df.columns:
            star_df.drop('dist_arcsec', inplace=True, axis=1)
        if 'ld_a' in star_df.columns:
            star_df['ld_a'] = star_df['ld_a'].fillna(0.25)
        if 'ld_a' in star_df.columns:
            star_df['ld_b'] = star_df['ld_b'].fillna(0.25)
        if 'st_teff' in target_row and not np.isnan(target_row['st_teff']):
            star_df['Teff'] = target_row['st_teff']
        elif 'koi_steff' in target_row and not np.isnan(target_row['koi_steff']):
            star_df['Teff'] = target_row['koi_steff']
        elif 'tce_steff' in target_row and not np.isnan(target_row['tce_steff']):
            star_df['Teff'] = target_row['tce_steff']
        star_df['Teff'] = star_df['Teff'].fillna(6000)
        if 'st_rad' in target_row and not np.isnan(target_row['st_rad']):
            star_df['radius'] = target_row['st_rad']
        elif 'koi_srad' in target_row and not np.isnan(target_row['koi_srad']):
            star_df['radius'] = target_row['koi_srad']
        elif 'tce_sradius' in target_row and not np.isnan(target_row['tce_sradius']):
            star_df['radius'] = target_row['tce_sradius']
        if 'lum' in star_df.columns:
            star_df['lum'] = star_df['lum'].fillna(1)
        if 'logg' in star_df.columns:
            star_df['logg'] = star_df['logg'].fillna(1)
        if 'v' in star_df.columns:
            star_df['v'] = star_df['v'].fillna(-10)
        if 'j' in star_df.columns:
            star_df['j'] = star_df['j'].fillna(-10)
        if 'k' in star_df.columns:
            star_df['k'] = star_df['k'].fillna(-10)
        if 'h' in star_df.columns:
            star_df['h'] = star_df['h'].fillna(-10)
        star_df['Teff'] = star_df['Teff'] / 60000 * teff_mod
        star_df['Teff'] = star_df['Teff'] if star_df['Teff'].iloc[0] <= 1 else 1
        star_df['radius'] = star_df['radius'] / 3 * rad_mod
        star_df['radius'] = star_df['radius'] if star_df['radius'].iloc[0] <= 1 else 1
        if 'mass' in star_df:
            star_df['mass'] = star_df['mass'] / 3 * mass_mod
            star_df['mass'] = star_df['mass'] if star_df['mass'].iloc[0] <= 1 else 1
        if 'lum' in star_df.columns:
            star_df['lum'] = sigmoid(star_df['lum'])
        if 'logg' in star_df.columns:
            star_df['logg'] = sigmoid(star_df['logg'])
        if 'v' in star_df.columns:
            star_df['v'] = (30 - star_df['v']) / 40
        if 'j' in star_df.columns:
            star_df['j'] = (30 - star_df['j']) / 40
        if 'k' in star_df.columns:
            star_df['k'] = (30 - star_df['k']) / 40
        if 'h' in star_df.columns:
            star_df['h'] = (30 - star_df['h']) / 40
        return star_df.iloc[0], ra, dec

    def _prepare_input_neighbour_stars(self, star_df):
        search_radius = 10
        pixel_size = 20.25
        max_distance = search_radius * pixel_size
        # Limiting the set to 15 stars, removing the first one, which is the target star
        star_df = star_df.iloc[1:16, :].copy()
        if len(star_df) > 0:
            star_df['Teff'] = star_df['Teff'].fillna(6000)
            star_df['radius'] = star_df['radius'].fillna(1)
            star_df['mass'] = star_df['mass'].fillna(1)
            star_df['lum'] = star_df['lum'].fillna(1)
            star_df['v'] = star_df['v'].fillna(-10)
            star_df['j'] = star_df['j'].fillna(-10)
            star_df['k'] = star_df['k'].fillna(-10)
            star_df['h'] = star_df['h'].fillna(-10)
            star_df['Teff'] = star_df['Teff'] / 40000
            star_df['radius'] = star_df['radius'] / 3
            star_df['radius'] = star_df['radius'] if star_df['radius'].iloc[0] <= 1 else 1
            star_df['mass'] = star_df['mass'] / 3
            star_df['mass'] = star_df['mass'] if star_df['mass'].iloc[0] <= 1 else 1
            star_df['lum'] = star_df['lum'] / 1000
            star_df['v'] = (30 - star_df['v']) / 40
            star_df['j'] = (30 - star_df['j']) / 40
            star_df['k'] = (30 - star_df['k']) / 40
            star_df['h'] = (30 - star_df['h']) / 40
            if 'dist_arcsec' in star_df.columns:
                star_df['dist_arcsec'] = star_df['dist_arcsec'] / max_distance
                star_df['dist_arcsec'] = star_df['dist_arcsec'].fillna(max_distance)
        return star_df

    @staticmethod
    def compute_snr(file_prefix, time, flux, duration, period, epoch, oot_range=5, fold=False, baseline=1, zero_epsilon=1e-7):
        duration_to_period = duration / period
        lc_df = pd.DataFrame(columns=['time', 'time_folded', 'flux'])
        lc_df['time'] = time
        lc_df['flux'] = flux
        if fold:
            lc_df['time_folded'] = foldedleastsquares.fold(lc_df['time'].to_numpy(), period, epoch + period / 2)
            lc_df = lc_df.sort_values(by=['time_folded'], ascending=True)
        else:
            lc_df['time_folded'] = lc_df['time']
        lc_it = lc_df[(lc_df['time_folded'] > 0.5 - duration_to_period / 2) &
                      (lc_df['time_folded'] < 0.5 + duration_to_period / 2)]
        lc_oot = lc_df[((lc_df['time_folded'] < 0.5 - duration_to_period) & (
                    lc_df['time_folded'] > 0.5 - duration_to_period * oot_range / 2)) |
                       ((lc_df['time_folded'] > 0.5 + duration_to_period) & (
                                   lc_df['time_folded'] < 0.5 + duration_to_period * oot_range / 2))]
        snr = (baseline - lc_it['flux'].mean()) * np.sqrt(len(lc_it)) / lc_oot['flux'].std()
        if np.isnan(snr):
            logging.info(f"NaN SNR for {file_prefix}")
            snr = zero_epsilon
        elif snr >= 20 - zero_epsilon:
            snr = 1 - zero_epsilon
        elif snr < -20:
            snr = zero_epsilon
        else:
            snr = (snr + 20) / 40 + zero_epsilon
        return snr

    def compute_og_df(self, read_og_df, object_id, duration):
        if len(read_og_df) > 0:
            if read_og_df['core_flux'].isnull().values.all():
                logging.info("Nan values found for CORE_OG for %s, zeroing OG", object_id)
                read_og_df.loc[read_og_df['core_flux'].isnull(), 'core_flux'] = 1
            if read_og_df['halo_flux'].isnull().values.all():
                logging.info("Nan values found for HALO_OG for %s, zeroing OG", object_id)
                read_og_df.loc[read_og_df['halo_flux'].isnull(), 'halo_flux'] = 1
            read_og_df['core_flux'], _ = LcbuilderHelper.detrend(read_og_df['time'].to_numpy(),
                                                              read_og_df['core_flux'].to_numpy(), duration * 4,
                                                              check_cadence=True)
            read_og_df['halo_flux'], _ = LcbuilderHelper.detrend(read_og_df['time'].to_numpy(),
                                                              read_og_df['halo_flux'].to_numpy(), duration * 4,
                                                              check_cadence=True)
            read_og_df['og_flux'] = read_og_df['halo_flux'] - read_og_df['core_flux']
        read_og_df = read_og_df.drop(columns=['halo_flux', 'core_flux'])
        return read_og_df

    @abstractmethod
    def class_weights(self):
        pass


# df = pd.read_csv("/mnt/DATA-2/ete6/injected_objects.csv", index_col=False)
# df = shuffle(df)
# df.reset_index(inplace=True, drop=True)
# img = IatsonModelGenerator(df, "/mnt/DATA-2/ete6/lcs", 20, [11, 2500, 500, 500, 500, 500, 500, 500])
# img.__getitem__(20)



from curve_prioritizer.curve_prioritizer_generator import CurvePrioritizerGenerator

model = 1 # keras model
batch_size = 100
injected_objects_df = pd.read_csv("injected_objects_combined.csv")
# TODO split in training/validation/test dfs
injected_objects_training_df = 1
curve_initializer_generator = CurvePrioritizerGenerator(injected_objects_training_df, batch_size=batch_size)
curve_initializer_generator_val = CurvePrioritizerGenerator(injected_objects_validation_df, batch_size=batch_size)
fit_history = model.fit(x=curve_initializer_generator,
                           steps_per_epoch=steps_per_epoch,
                           epochs=epochs, verbose=1,
                           validation_data=curve_initializer_generator_val,
                           callbacks=hyperparams.callbacks,
                           use_multiprocessing=cores > 0, cores)


import foldedleastsquares
import keras
import numpy as np
import pandas as pd


class LcDataInputGenerator(keras.utils.Sequence):

    def __init__(self, lc_filenames, labels, batch_size, lc_len, mode="short"):
        self.lc_filenames = lc_filenames
        self.labels = labels
        self.batch_size = batch_size
        self.lc_len = lc_len

    def __len__(self):
        return (np.ceil(len(self.lc_filenames) / float(self.batch_size))).astype(np.int)

    def __getitem__(self, idx):
        batch_x = self.lc_filenames[idx * self.batch_size: (idx + 1) * self.batch_size]
        batch_y = self.labels[idx * self.batch_size: (idx + 1) * self.batch_size]
        # TODO inputs: flux (2500, 7)
        # TODO inputs: centroidsbck (2500, 5)

        # TODO inputs: global centroids bck  (2500, 5)
        # TODO inputs: glbal flux (2500, 7)
        # TODO inputs: focus odd harmonic centroids (500, 5)
        # TODO inputs: focus odd subharmonic centroids (500, 5)
        # TODO inputs: focus odd centroids (500, 5)
        # TODO inputs: focus even harmonic centroids (500, 5)
        # TODO inputs: focus even subharmonic centroids (500, 5)
        # TODO inputs: stellar model (6, 1)
        # TODO inputs: stellar model (6, 1)
        # TODO inputs: stellar model (6, 1)
        # TODO inputs: stellar model (6, 1)
        # TODO inputs: stellar model (6, 1)
        # TODO inputs: stellar model (6, 1)
        for lc_filename in self.lc_filenames:
            star_df = pd.read_csv(lc_filename + "/params_star.csv")
            star_input = [star_df.iloc[0]["h"], star_df.iloc[0]["j"], star_df.iloc[0]["k"], star_df.iloc[0]["v"],
                          star_df.iloc[0]["R_star"], star_df.iloc[0]["M_star"], star_df.iloc[0]["Teff_star"],
                          star_df.iloc[0]["logg"], star_df.iloc[0]["feh"],
                          star_df.iloc[0]["ld_a"], star_df.iloc[0]["ld_b"]]
            # TODO normalize star_input values
            # TODO check for empty star_input values and select replacement
            flux_input = []
            lc_data = pd.read_csv(lc_filename + "/lc_data_" + self.mode + ".csv")
            if "_tp_" in lc_filename:
                ois_df = pd.read_csv(lc_filename + "/ois.csv")
                oi_index = int(lc_filename.split("_tp_")[1])
                period = ois_df.iloc[oi_index]["period"]
                t0 = ois_df.iloc[oi_index]["t0"]
            elif "_fp_" in lc_filename:
                ois_df = pd.read_csv(lc_filename + "/ois.csv")
                oi_index = int(lc_filename.split("_fp_")[1])
                period = ois_df.iloc[oi_index]["period"]
                t0 = ois_df.iloc[oi_index]["t0"]
            else:
                min_time = np.nanmin(lc_data["time"])
                period = np.random.uniform(size=1, low=0.5, high=20)
                t0 = np.random.uniform(size=1, low=min_time, high=min_time + period)
            lc_data["time_folded_period"] = foldedleastsquares.fold(lc_data["time"].to_numpy(), period, T0=t0)
            lc_data["time_folded_period"] = foldedleastsquares.fold(lc_data["time"].to_numpy(), period, T0=t0)
            lc_data["time_folded_period"] = foldedleastsquares.fold(lc_data["time"].to_numpy(), period, T0=t0)
            lc_data["time_folded_period"] = foldedleastsquares.fold(lc_data["time"].to_numpy(), period, T0=t0)
            lc_data = lc_data.sort_values(by=['time'], ascending=True)
            # Discover, visualize, and preprocess data using pandas if needed.

            data = data.to_numpy()
        return np.array([
            resize(imread('/content/all_images/' + str(file_name)), (80, 80, 3))
            for file_name in batch_x]) / 255.0, np.array(batch_y)


import logging

from exoml.ml.encoding.time_position import value_encode_times
from sklearn.utils import shuffle
import tensorflow as tf
import numpy as np
import pandas as pd


class HaunterModelGenerator(tf.keras.utils.Sequence):
    """
    Sequence generator for DETREND model batches
    """
    def __init__(self, dataset, batch_size, input_size, zero_epsilon=1e-7, shuffle=True):
        self.zero_epsilon = zero_epsilon
        self.dataset = dataset
        self.batch_size = batch_size
        self.input_size = input_size
        self.shuffle = shuffle
        if self.shuffle:
            self.dataset = self.dataset.sample(frac=1).reset_index(drop=True)

    def __len__(self):
        return (np.ceil(len(self.dataset) / float(self.batch_size))).astype(int)

    def __getitem__(self, idx):
        batch_rows = self.dataset[idx * self.batch_size: (idx + 1) * self.batch_size]
        batch_data_array = batch_rows.loc[:, ['mass_Mearth_input_norm', 'CMF_input_norm', 'Zenv_input_norm',
                                              'Zwater_core_input_norm', 'Tsurf_K_input_norm', 'Psurf_bar_input_norm']]
        batch_data_values = batch_rows.loc[:, ['radius_Rearth_output_norm', 'entropy_SI_output_norm',
                                               'f_s_SI_output_norm', 'mass_Mearth_input_norm', 'CMF_input_norm', 'Zenv_input_norm',
                                              'Zwater_core_input_norm', 'Tsurf_K_input_norm', 'Psurf_bar_input_norm']]
        #assert np.all((batch_data_array > 0) & (batch_data_array < 1))
        #assert np.all((batch_data_values > 0) & (batch_data_values < 1))
        return batch_data_array.to_numpy(dtype=np.float32), batch_data_values.to_numpy(dtype=np.float32)

    def class_weights(self):
        return None



from keras.callbacks import Callback


class GetWeights(Callback):
    # Keras callback which collects values of weights and biases at each epoch
    def __init__(self, model_weights_logger, save_dict=False):
        super(GetWeights, self).__init__()
        self.weight_dict = {}
        self.save_dict = save_dict
        self.model_weights_logger = model_weights_logger

    def on_batch_end(self, batch, logs=None):
        self.model_weights_logger.log_model_weights(self.model)



import logging
import os
import sys

import matplotlib.pyplot as plt
import foldedleastsquares
from lcbuilder.helper import LcbuilderHelper
from numpy.random import default_rng
from exoml.ete6.ete6_generator import Ete6ModelGenerator
from exoml.ml.encoding.time_position import value_encode_times
from sklearn.utils import shuffle
import numpy as np
import pandas as pd


class IatsonOgModelGenerator(Ete6ModelGenerator):
    def __init__(self, injected_objects_df, lcs_dir, batch_size, input_sizes, type_to_label, zero_epsilon=1e-7,
                 measurements_per_point=2, plot_inputs=False, fixed_target_id=None, store_arrays=False,
                 from_arrays=True, shuffle_batch=True):
        super().__init__(zero_epsilon, shuffle_batch=shuffle_batch)
        self.injected_objects_df = injected_objects_df
        self.kics_lcs_dir = lcs_dir + '/q1_q17/lcs/'
        self.tics_lcs_dir = lcs_dir + '/ete6/lcs/'
        self.batch_size = batch_size
        self.input_sizes = input_sizes
        self.random_number_generator = default_rng()
        self.measurements_per_point = measurements_per_point
        self.type_to_label = type_to_label
        self.plot_inputs = plot_inputs
        self.fixed_target_id = fixed_target_id
        self.store_arrays = store_arrays
        self.from_arrays = from_arrays

    def __len__(self):
        return (np.ceil(len(self.injected_objects_df) / float(self.batch_size))).astype(int)

    def class_weights(self):
        return {0: 1, 1: 1}

    def _plot_df(self, df, type, scenario):
        fig, axs = plt.subplots(2, 3, figsize=(12, 6), constrained_layout=True)
        axs[0][0].scatter(df['#time'], df['flux'])
        axs[0][1].scatter(df['#time'], df['centroid_x'])
        axs[0][2].scatter(df['#time'], df['centroid_y'])
        axs[1][0].scatter(df['#time'], df['bck_flux'])
        axs[1][1].scatter(df['#time'], df['motion_x'])
        axs[1][2].scatter(df['#time'], df['motion_y'])
        plt.title(type + " " + scenario)
        plt.show()
        plt.clf()
        plt.close()

    def _plot_input(self, input_array, input_err_array, type, scenario):
        if self.plot_inputs:
            transposed_err_array = []
            transposed_array = np.transpose(input_array)
            fig, axs = plt.subplots(2, 2, figsize=(24, 12), constrained_layout=True)
            current_array = transposed_array[0]
            current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            current_array = current_array[current_array_mask]
            time_array = current_array
            axs[0][0].scatter(np.arange(0, len(current_array)), current_array)
            # axs[0][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            axs[0][0].set_title("Time")
            current_array = transposed_array[1]
            current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            current_array = current_array[current_array_mask]
            if input_err_array is not None:
                transposed_err_array = np.transpose(input_err_array)
                axs[0][1].errorbar(time_array[current_array_mask], current_array, ls='none',
                                   yerr=transposed_err_array[1][current_array_mask], color="orange", alpha=0.5)
            axs[0][1].scatter(time_array[current_array_mask], current_array)
            if len(transposed_array) > 2:
                current_array = transposed_array[2]
                current_array_mask = np.argwhere(
                    (~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
                current_array = current_array[current_array_mask]
                if input_err_array is not None:
                    axs[1][0].errorbar(time_array[current_array_mask], transposed_array[2], ls='none',
                                       yerr=transposed_err_array[2][current_array_mask], color="orange", alpha=0.5)
                axs[1][0].scatter(time_array[current_array_mask], current_array)
            #axs[0][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            axs[0][1].set_title("Flux")
            axs[1][0].set_title("Flux 1")
            # current_array = transposed_array[8]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[1][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[8][current_array_mask], color="orange", alpha=0.5)
            # axs[1][1].scatter(time_array[current_array_mask], current_array)
            # # axs[1][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[1][1].set_title("Flux 2")
            # current_array = transposed_array[9]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[2][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[9][current_array_mask], color="orange", alpha=0.5)
            # axs[2][0].scatter(time_array[current_array_mask], current_array)
            # # axs[2][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[2][0].set_title("Flux 3")
            # current_array = transposed_array[10]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[2][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[10][current_array_mask], color="orange", alpha=0.5)
            # axs[2][1].scatter(time_array[current_array_mask], current_array)
            # # axs[2][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[2][1].set_title("Flux 4")
            # current_array = transposed_array[11]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[3][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[11][current_array_mask], color="orange", alpha=0.5)
            # axs[3][0].scatter(time_array[current_array_mask], current_array)
            # # axs[3][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array),
            # #                    np.mean(current_array) + 6 * np.std(current_array))
            # axs[3][0].set_title("Flux 5")
            # current_array = transposed_array[6]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[3][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[6][current_array_mask], color="orange", alpha=0.5)
            # axs[3][1].scatter(time_array[current_array_mask], current_array)
            # # axs[3][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[3][1].set_title("Bck Flux")
            # current_array = transposed_array[2]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[4][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[2][current_array_mask], color="orange", alpha=0.5)
            # axs[4][0].scatter(time_array[current_array_mask], current_array)
            # # axs[4][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[4][0].set_title("Centroid X")
            # current_array = transposed_array[3]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[4][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[3][current_array_mask], color="orange", alpha=0.5)
            # axs[4][1].scatter(time_array[current_array_mask], current_array)
            # # axs[4][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[4][1].set_title("Centroid Y")
            # current_array = transposed_array[4]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[5][0].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[4][current_array_mask], color="orange", alpha=0.5)
            # axs[5][0].scatter(time_array[current_array_mask], current_array)
            # # axs[5][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[5][0].set_title("Motion Y")
            # current_array = transposed_array[5]
            # current_array_mask = np.argwhere((~np.isnan(current_array)) & (current_array > self.zero_epsilon)).flatten()
            # current_array = current_array[current_array_mask]
            # axs[5][1].errorbar(time_array[current_array_mask], current_array,
            #                    yerr=transposed_err_array[5][current_array_mask], color="orange", alpha=0.5)
            # axs[5][1].scatter(time_array[current_array_mask], current_array)
            # # axs[5][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
            # axs[5][1].set_title("Motion Y")
            fig.suptitle(type + " " + scenario)
            plt.show()
            plt.clf()
            plt.close()

    def plot_single_data(self, lc_df, target_row):
        fig, axs = plt.subplots(1, 1, figsize=(8, 4), constrained_layout=True)
        axs.scatter(lc_df['#time'], lc_df['flux_0'])
        axs.set_ylim(0.5 - target_row['tce_depth'] / 0.5e6, 0.505)
        plt.show()
        plt.clf()
        plt.close()

    def __getitem__(self, idx):
        max_index = (idx + 1) * self.batch_size
        max_index = len(self.injected_objects_df) if max_index > len(self.injected_objects_df) else max_index
        target_indexes = np.arange(idx * self.batch_size, max_index, 1)
        injected_objects_df = self.injected_objects_df.iloc[target_indexes]
        if self.shuffle_batch:
            injected_objects_df = shuffle(injected_objects_df)
        folded_og_array = np.empty((len(target_indexes), self.input_sizes[3], 2))
        folded_og_array_err = np.empty((len(target_indexes), self.input_sizes[3], 2))
        batch_data_values = np.empty((len(target_indexes), 1))
        i = 0
        for df_index, target_row in injected_objects_df.iterrows():
            if self.fixed_target_id is not None:
                target_row = self.injected_objects_df[self.injected_objects_df['object_id'] == self.fixed_target_id[0]]
                target_row = target_row[target_row['period'] == self.fixed_target_id[1]]
            #TODO refactor to use mission for object_id
            object_id = target_row['object_id'].split(' ')
            mission_id = object_id[0]
            target_id = int(object_id[1])
            period = target_row['period']
            epoch = target_row['epoch']
            duration = target_row['duration(h)'] / 24
            duration_to_period = duration / period
            type = target_row['type']
            multitype = target_row['multitype']
            batch_data_values[i] = [0.0 if isinstance(multitype, str) and 'tce_og' in multitype else 1.0]
            lcs_dir = self.kics_lcs_dir if mission_id == 'KIC' else self.tics_lcs_dir
            file_prefix = lcs_dir + '/' + mission_id + '_' + str(target_id) + '_' + str(round(period, 2)) \
                if mission_id == 'KIC' else lcs_dir + '/' + mission_id + ' ' + str(target_id) + '_' + str(round(period, 2))
            if self.from_arrays:
                folded_og_array[i] = np.loadtxt(file_prefix + '_input_og.csv', delimiter=',')
                folded_og_array_err[i] = np.loadtxt(file_prefix + '_input_og_err.csv', delimiter=',')
            else:
                #TODO refactor to use mission for _lc files
                og_short_filename = file_prefix + '_og_short.csv'
                og_long_filename = file_prefix + '_og_long.csv'
                og_df = None
                if os.path.exists(og_long_filename):
                    read_og_df = pd.read_csv(og_long_filename, usecols=['time', 'og_flux', 'halo_flux', 'core_flux'],
                                             low_memory=True)
                    read_og_df = self.compute_og_df(read_og_df, object_id, duration)
                    og_df = read_og_df if og_df is None else og_df.append(read_og_df, ignore_index=True)
                if os.path.exists(og_short_filename):
                    read_og_df = pd.read_csv(og_short_filename, usecols=['time', 'og_flux', 'halo_flux', 'core_flux'],
                                             low_memory=True)
                    read_og_df = self.compute_og_df(read_og_df, object_id, duration)
                    og_df = read_og_df if og_df is None else og_df.append(read_og_df, ignore_index=True)
                og_df['time'] = self.fold(og_df['time'].to_numpy(), period, epoch + period / 2)
                og_df = og_df.sort_values(by=['time'])
                og_df = self._prepare_input_og(og_df)
                og_df = og_df[(og_df['time'] > 0.5 - duration_to_period * 3) & (
                        og_df['time'] < 0.5 + duration_to_period * 3)]
                folded_og_array[i], folded_og_array_err[i] = self.bin_by_time(og_df.to_numpy(), self.input_sizes[3], target_row['object_id'])
            self.assert_in_range(object_id, folded_og_array[i], folded_og_array_err[i])
            self._plot_input(folded_og_array[i], folded_og_array_err[i], target_row['object_id'] + "_" + type, "OG")
            if self.store_arrays:
                logging.info("Storing arrays into prefix " + file_prefix)
                np.savetxt(file_prefix + '_input_og.csv', folded_og_array[i], delimiter=',')
                np.savetxt(file_prefix + '_input_og_err.csv', folded_og_array_err[i], delimiter=',')
            i = i + 1
        return folded_og_array, batch_data_values
